import unittest
import os
from baiji import s3

class TestS3Path(unittest.TestCase):

    def test_s3_parse(self):
        def test(s, expect, on_windows=None):
            if on_windows and os.name == 'nt':
                expect = expect[:2] + (on_windows,)
            o = s3.path.parse(s)
            for ii in range(3):
                self.assertEqual(o[ii], expect[ii])

        test('s3://foo/path/to/bar', ('s3', 'foo', '/path/to/bar'))
        test('s3://xx', ('s3', 'xx', ''))
        test('s3://xx/', ('s3', 'xx', '/'))
        self.assertRaises(ValueError, s3.path.parse, 's3:///path/to/bar')
        test('file:///path/to/blah', ('file', '', '/path/to/blah'), on_windows=r'C:\path\to\blah')
        test('/path/to/blah', ('file', '', '/path/to/blah'), on_windows=r'C:\path\to\blah')
        test('path/to/blah', ('file', '', os.path.abspath('./path/to/blah')))
        test('~/path/to/blah', ('file', '', os.path.abspath(os.path.expanduser('~/path/to/blah'))))
        test('', ('file', '', os.path.abspath('.')))

        self.assertTrue(s3.path.islocal('/path/to/bar'))
        self.assertFalse(s3.path.islocal('s3://foo/path/to/bar'))
        self.assertFalse(s3.path.isremote('/path/to/bar'))
        self.assertTrue(s3.path.isremote('s3://foo/path/to/bar'))

        self.assertEqual(s3.path.join('s3://foo/path', 'to/bar'), 's3://foo/path/to/bar')
        if os.name == 'nt':
            self.assertEqual(s3.path.join('/foo/path', 'to/bar'), r'C:\foo\path\to\bar')
        else:
            self.assertEqual(s3.path.join('/foo/path', 'to/bar'), '/foo/path/to/bar')

        self.assertEqual(s3.path.basename('file:///foo/path/to/bar.baz'), 'bar.baz')
        self.assertEqual(s3.path.basename('s3://bucket/foo/path/to/bar.baz'), 'bar.baz')

        test('C:\\path\\of\\quux', ('file', '', 'C:\\path\\of\\quux'))
        test('file:///C:\\path\\of\\quux', ('file', '', 'C:\\path\\of\\quux'))

        with self.assertRaises(ValueError):
            s3.path.parse('file:///foo;bar')

        # Make sure it preserves trailing slashes
        test('path/to/blah/', ('file', '', os.path.join(os.path.abspath('./path/to/blah'), '')))
        test('s3://foo/path/to/blah/', ('s3', 'foo', '/path/to/blah/'))



def mkdir_p(path):
    '''
    Implements ``mkdir -p``.
    '''
    import errno
    import os
    if len(path) > 0:
        try:
            os.makedirs(path)
        except OSError as ex:
            if ex.errno == errno.EEXIST:
                pass
            else: raise



import os
from baiji.exceptions import AWSCredentialsMissing

class Credentials(object):
    '''
    Amazon AWS credential object

    If created without an explicit path to the credential file, it will use the environment variable
    ``BODYLABS_CREDENTIAL_FILE``, or if that is not set, default to ``~/.bodylabs``.

    The file is to be a yaml file containing a dict with the keys ``AWS_ACCESS_KEY`` and ``AWS_SECRET``.

    If ``~/.bodylabs`` doesn't exist, we will try to read in the aws cli config file, ``~/.aws/credentials``.

    If the keys are set via environment variables, these will override anything set in a file.

    The credential object makes these availiable as ``o.key`` and ``o.secret``.
    '''
    environment_variable = 'BODYLABS_CREDENTIAL_FILE'
    default_path = '~/.bodylabs'
    aws_cli_path = '~/.aws/credentials'

    def __init__(self):
        self._raw_data = None

    def load(self):
        from baiji.util import yaml
        from baiji.util.environ import getenvpath

        if self._raw_data is not None:
            return self._raw_data

        path = getenvpath(self.environment_variable, self.default_path)
        if os.path.isfile(path):
            try:
                self._raw_data = yaml.load(path)
                return self._raw_data
            except IOError:
                raise AWSCredentialsMissing("Unable to read AWS configuration file: {}".format(path))
        elif os.path.isfile(os.path.expanduser(self.aws_cli_path)):
            import ConfigParser
            aws_cli_config = ConfigParser.ConfigParser()
            aws_cli_config.read([os.path.expanduser(self.aws_cli_path)])
            self._raw_data = {
                'AWS_ACCESS_KEY': aws_cli_config.get('default', 'aws_access_key_id'),
                'AWS_SECRET': aws_cli_config.get('default', 'aws_secret_access_key'),
            }
            return self._raw_data
        else:
            raise AWSCredentialsMissing("Unable to read AWS configuration file: {}".format(path))

    def _try(self, var_list, key):
        for var in var_list:
            val = os.getenv(var, None)
            if val:
                return val
        try:
            return self.load()[key]
        except KeyError:
            raise AWSCredentialsMissing('AWS configuration is missing or ill formed.')

    @property
    def key(self):
        return self._try(['AWS_ACCESS_KEY_ID', 'AWS_ACCESS_KEY'], 'AWS_ACCESS_KEY')
    @property
    def secret(self):
        return self._try(['AWS_SECRET_ACCESS_KEY', 'AWS_SECRET'], 'AWS_SECRET')

credentials = Credentials()

def is_available():
    from baiji.util.reachability import internet_reachable
    try:
        credentials.load()
    except AWSCredentialsMissing:
        return False
    return internet_reachable()



import uuid
import os
import mock
from baiji import s3
from baiji.test_s3 import TestAWSBase

class TestCachedFile(TestAWSBase):
    def setUp(self):
        super(TestCachedFile, self).setUp()
        s3.cp(self.local_file, self.remote_file("openable"))
        with open(self.local_file, 'rb') as f:
            self.truth = f.read()

    def test_s3_open_read_remote_file_with_context_manager(self):
        self.assert_s3_exists(self.remote_file("openable"))

        with s3.open(self.remote_file("openable"), 'r') as f:
            tempname = f.name
            self.assertEqual(self.truth, f.read())

        self.assertFalse(os.path.exists(tempname))

    def test_s3_open_read_remote_file_without_context_manager(self):
        self.assert_s3_exists(self.remote_file("openable"))

        f = s3.open(self.remote_file("openable"), 'r')
        tempname = f.name
        self.assertEqual(self.truth, f.read())
        f.close()

        self.assertFalse(os.path.exists(tempname))

    def test_s3_open_read_local_file_with_context_manager(self):
        self.assert_s3_exists(self.local_file)

        with s3.open(self.local_file, 'r') as f:
            self.assertEqual(self.truth, f.read())
            self.assertEqual(f.name, self.local_file)

    def test_s3_open_read_local_file_without_context_manager(self):
        self.assert_s3_exists(self.local_file)

        f = s3.open(self.local_file, 'r')
        self.assertEqual(self.truth, f.read())
        self.assertEqual(f.name, self.local_file)
        f.close()

    def test_s3_open_write_remote_file_with_context_manager(self):
        remote_file_name = self.remote_file("write_test_1")
        local_file_name = os.path.join(self.tmp_dir, "write_test_1")
        self.assert_s3_does_not_exist(remote_file_name)

        with s3.open(remote_file_name, 'w') as f:
            tempname = f.name
            f.write(self.truth)

        self.assertFalse(os.path.exists(tempname))
        self.assert_s3_exists(remote_file_name)
        # download and confirm that it contains the correct contents
        s3.cp(remote_file_name, local_file_name)
        with open(local_file_name) as f:
            self.assertEqual(self.truth, f.read())

    def test_s3_open_write_local_file_creates_parent_dirs(self):
        local_file_name = os.path.join(self.tmp_dir, "subdir", "write_test_subdir")
        self.assert_s3_does_not_exist(local_file_name)
        self.assert_s3_does_not_exist(os.path.dirname(local_file_name))

        with s3.open(local_file_name, 'w') as f:
            f.write(self.truth)

        self.assert_s3_exists(local_file_name)
        self.assert_s3_exists(os.path.dirname(local_file_name))

    def test_s3_open_write_remote_file_without_context_manager(self):
        remote_file_name = self.remote_file("write_test_2")
        local_file_name = os.path.join(self.tmp_dir, "write_test_2")
        self.assert_s3_does_not_exist(remote_file_name)

        f = s3.open(remote_file_name, 'w')
        tempname = f.name
        f.write(self.truth)
        f.close()

        self.assertFalse(os.path.exists(tempname))
        self.assert_s3_exists(remote_file_name)
        # download and confirm that it contains the correct contents
        s3.cp(remote_file_name, local_file_name)
        with open(local_file_name) as f:
            self.assertEqual(self.truth, f.read())

    def test_s3_open_write_local_file_with_context_manager(self):
        local_file_name = os.path.join(self.tmp_dir, "write_test_local_1")
        self.assert_s3_does_not_exist(local_file_name)

        with s3.open(local_file_name, 'w') as f:
            self.assertEqual(f.name, local_file_name)
            f.write(self.truth)

        self.assert_s3_exists(local_file_name)
        # confirm that it contains the correct contents
        with open(local_file_name) as f:
            self.assertEqual(self.truth, f.read())

    def test_s3_open_write_local_file_without_context_manager(self):
        local_file_name = os.path.join(self.tmp_dir, "write_test_local_2")
        self.assert_s3_does_not_exist(local_file_name)

        f = s3.open(local_file_name, 'w')
        self.assertEqual(f.name, local_file_name)
        f.write(self.truth)
        f.close()

        self.assert_s3_exists(local_file_name)
        # confirm that it contains the correct contents
        with open(local_file_name) as f:
            self.assertEqual(self.truth, f.read())

    def test_s3_open_exclusive_write_remote_file_throws_error_for_existing_file(self):
        self.assert_s3_exists(self.remote_file("openable"))
        with self.assertRaises(s3.KeyExists):
            s3.open(self.remote_file("openable"), 'x')

    def test_s3_open_exclusive_write_local_file_throws_error_for_existing_file(self):
        self.assert_s3_exists(self.local_file)
        with self.assertRaises(s3.KeyExists):
            s3.open(self.local_file, 'x')

    def test_s3_open_exclusive_write_remote_file_throws_error_for_double_open(self):
        self.assert_s3_does_not_exist(self.remote_file("exclusive_write_test_3"))
        with s3.open(self.remote_file("exclusive_write_test_3"), 'x'):
            with self.assertRaises(s3.KeyExists):
                with s3.open(self.remote_file("exclusive_write_test_3"), 'x'):
                    pass

    def test_s3_open_exclusive_write_local_file_throws_error_for_double_open(self):
        local_file_name = os.path.join(self.tmp_dir, "exclusive_write_test_local_3")
        self.assert_s3_does_not_exist(local_file_name)
        with s3.open(local_file_name, 'x'):
            with self.assertRaises(s3.KeyExists):
                with s3.open(local_file_name, 'x'):
                    pass

    def test_s3_open_exclusive_write_remote_file_with_context_manager(self):
        remote_file_name = self.remote_file("exclusive_write_test_1")
        local_file_name = os.path.join(self.tmp_dir, "exclusive_write_test_1")
        self.assert_s3_does_not_exist(remote_file_name)

        with s3.open(remote_file_name, 'x') as f:
            self.assert_s3_exists(remote_file_name)
            tempname = f.name
            f.write(self.truth)

        self.assertFalse(os.path.exists(tempname))
        self.assert_s3_exists(remote_file_name)
        # download and confirm that it contains the correct contents
        s3.cp(remote_file_name, local_file_name)
        with open(local_file_name) as f:
            self.assertEqual(self.truth, f.read())

    def test_s3_open_exclusive_write_remote_file_without_context_manager(self):
        remote_file_name = self.remote_file("exclusive_write_test_2")
        local_file_name = os.path.join(self.tmp_dir, "exclusive_write_test_2")
        self.assert_s3_does_not_exist(remote_file_name)

        f = s3.open(remote_file_name, 'x')
        self.assert_s3_exists(remote_file_name)
        tempname = f.name
        f.write(self.truth)
        f.close()

        self.assertFalse(os.path.exists(tempname))
        self.assert_s3_exists(remote_file_name)
        # download and confirm that it contains the correct contents
        s3.cp(remote_file_name, local_file_name)
        with open(local_file_name) as f:
            self.assertEqual(self.truth, f.read())

    def test_s3_open_exclusive_write_local_file_with_context_manager(self):
        local_file_name = os.path.join(self.tmp_dir, "exclusive_write_test_local_1")
        self.assert_s3_does_not_exist(local_file_name)

        with s3.open(local_file_name, 'x') as f:
            self.assertEqual(f.name, local_file_name)
            f.write(self.truth)

        self.assert_s3_exists(local_file_name)
        # confirm that it contains the correct contents
        with open(local_file_name) as f:
            self.assertEqual(self.truth, f.read())

    def test_s3_open_exclusive_write_local_file_without_context_manager(self):
        local_file_name = os.path.join(self.tmp_dir, "exclusive_write_test_local_2")
        self.assert_s3_does_not_exist(local_file_name)

        f = s3.open(local_file_name, 'x')
        self.assertEqual(f.name, local_file_name)
        f.write(self.truth)
        f.close()

        self.assert_s3_exists(local_file_name)
        # confirm that it contains the correct contents
        with open(local_file_name) as f:
            self.assertEqual(self.truth, f.read())

    def test_s3_open_read_raises_error_for_nonexistent_remote_file(self):
        nonexistent_file = self.remote_file(str(uuid.uuid4()))
        self.assert_s3_does_not_exist(nonexistent_file)
        with self.assertRaises(s3.KeyNotFound):
            s3.open(nonexistent_file, 'r')

    def test_s3_open_write_does_not_raise_error_for_nonexistent_remote_file(self):
        nonexistent_file = self.remote_file(str(uuid.uuid4()))
        self.assert_s3_does_not_exist(nonexistent_file)
        with s3.open(nonexistent_file, 'w') as f:
            test_string = 'FOO!\n'
            f.write(test_string)
        self.assert_s3_exists(nonexistent_file)

    def test_s3_open_write_update_raises_error_for_nonexistent_remote_file(self):
        nonexistent_file = self.remote_file(str(uuid.uuid4()))
        self.assert_s3_does_not_exist(nonexistent_file)
        with self.assertRaises(NotImplementedError):
            s3.open(nonexistent_file, 'w+')

    def test_s3_open_exclusive_write_does_not_raise_error_for_nonexistent_remote_file(self):
        nonexistent_file = self.remote_file(str(uuid.uuid4()))
        self.assert_s3_does_not_exist(nonexistent_file)
        with s3.open(nonexistent_file, 'x') as f:
            test_string = 'FOO!\n'
            f.write(test_string)
        self.assert_s3_exists(nonexistent_file)

    def test_s3_open_read_raises_error_for_nonexistent_local_file(self):
        nonexistent_file = os.path.join(self.tmp_dir, "test_s3_open_read_raises_error_for_nonexistent_local_file")
        self.assert_s3_does_not_exist(nonexistent_file)
        with self.assertRaises(s3.KeyNotFound):
            s3.open(nonexistent_file, 'r')

    def test_s3_open_write_does_not_raise_error_for_nonexistent_local_file(self):
        nonexistent_file = os.path.join(self.tmp_dir, "test_s3_open_write_does_not_raise_error_for_nonexistent_local_file")
        self.assert_s3_does_not_exist(nonexistent_file)
        with s3.open(nonexistent_file, 'w') as f:
            test_string = 'FOO!\n'
            f.write(test_string)
        self.assert_s3_exists(nonexistent_file)

    def test_s3_open_write_update_does_not_raise_error_for_nonexistent_local_file(self):
        nonexistent_file = os.path.join(self.tmp_dir, "test_s3_open_write_update_does_not_raise_error_for_nonexistent_local_file")
        self.assert_s3_does_not_exist(nonexistent_file)
        with s3.open(nonexistent_file, 'w+') as f:
            test_string = 'FOO!\n'
            f.write(test_string)
            f.seek(0)
            self.assertEqual(test_string, f.read())
        self.assert_s3_exists(nonexistent_file)

    def test_s3_open_exclusive_write_does_not_raise_error_for_nonexistent_local_file(self):
        nonexistent_file = os.path.join(self.tmp_dir, "test_s3_open_exclusive_write_does_not_raise_error_for_nonexistent_local_file")
        self.assert_s3_does_not_exist(nonexistent_file)
        with s3.open(nonexistent_file, 'x') as f:
            test_string = 'FOO!\n'
            f.write(test_string)
        self.assert_s3_exists(nonexistent_file)

    def test_s3_open_local_underlying_error_raises_ioerror_with_errno_and_strerror(self):
        import errno
        with self.assertRaises(IOError) as ctx:
            s3.open(os.getcwd(), 'w')
        self.assertEquals(ctx.exception.errno, errno.EISDIR)
        self.assertIn('Is a directory:', ctx.exception.strerror)

    @mock.patch('baiji.s3.CachedFile.upload')
    def test_s3_open_write_calls_upload(self, upload):
        remote_file_name = self.remote_file("write_test_1")
        with s3.open(remote_file_name, 'w'):
            self.assertFalse(upload.called)
        self.assertTrue(upload.called)

    @mock.patch('baiji.s3.CachedFile.upload')
    def test_s3_open_write_does_not_upload_if_exception_raised(self, upload):
        remote_file_name = self.remote_file("write_test_1")
        try:
            with s3.open(remote_file_name, 'w'):
                raise AttributeError()
        except AttributeError:
            pass
        self.assertFalse(upload.called)

        # Sanity check
        with s3.open(remote_file_name, 'w'):
            pass
        self.assertTrue(upload.called)

    @mock.patch('baiji.s3.CachedFile.upload')
    def test_s3_open_read_does_not_call_upload(self, upload):
        self.assert_s3_exists(self.remote_file("openable"))
        with s3.open(self.remote_file("openable"), 'r'):
            pass
        self.assertFalse(upload.called)



class AWSError(Exception):
    pass

class AWSCredentialsMissing(AWSError):
    pass

class InvalidSchemeException(ValueError):
    pass

class S3Exception(AWSError):
    pass

class KeyNotFound(S3Exception):
    pass

class KeyExists(S3Exception):
    pass

class _TransientError(RuntimeError):
    pass

def get_transient_error_class():
    try:
        from guts.service.exceptions import TransientError  # That, precisely, is the point. pylint: disable=no-name-in-module
        return TransientError
    except ImportError:
        return _TransientError

class S3Warning(RuntimeWarning):
    pass

class EventualConsistencyWarning(S3Warning):
    pass



def random_data():
    def random_line():
        import random
        import string
        return ''.join(random.choice(string.ascii_letters) for x in range(80))
    return '\n'.join(random_line() for x in range(30))

def create_random_temporary_file():
    '''
    Create a temporary file with random contents, and return its path.

    The caller is responsible for removing the file when done.
    '''
    from baiji.util import tempfile
    with tempfile.NamedTemporaryFile('w', delete=False) as f:
        f.write(random_data())
        return f.name



import os
import unittest
import uuid

class EnvironGetEnvPath(unittest.TestCase):

    def test_that_getenvpath_reads_from_env(self):
        from baiji.util.environ import getenvpath
        env_var = str(uuid.uuid4())
        os.environ[env_var] = 'foo'
        self.assertEquals(getenvpath(env_var, 'bar'), os.path.abspath('foo'))

    def test_that_getenvpath_supports_default(self):
        from baiji.util.environ import getenvpath
        env_var = str(uuid.uuid4())
        self.assertEquals(getenvpath(env_var, 'bar'), os.path.abspath('bar'))

    def test_that_getenvpath_expands_user_home(self):
        from baiji.util.environ import getenvpath
        env_var = str(uuid.uuid4())
        os.environ[env_var] = '~/foo'
        self.assertEquals(os.environ.get(env_var, 'bar'), '~/foo')
        self.assertEquals(getenvpath(env_var, 'bar'), os.path.expanduser('~/foo'))



import os
import unittest
import shutil
import uuid
from test.test_support import EnvironmentVarGuard
from baiji.util import tempfile

class TestTempfile(unittest.TestCase):

    def test_that_NamedTemporaryFile_honors_TMP_env_var(self):
        env = EnvironmentVarGuard()
        test_temp_dir = 'test_tempfile_' + str(uuid.uuid4())
        env.set('BAIJI_TMP', test_temp_dir)
        with env:
            with tempfile.NamedTemporaryFile('w') as tf:
                self.assertEquals(os.path.dirname(tf.name), os.path.join(os.getcwd(), test_temp_dir))
        shutil.rmtree(test_temp_dir)



def getenvpath(env_var, default=None):
    '''
    A version of os.getenv that handles paths a bit better
    '''
    import os
    path = os.getenv(env_var, default)
    return os.path.realpath(os.path.expanduser(path))



import unittest
import uuid
from baiji import s3
from baiji.test_s3 import TEST_BUCKET

class TestS3TmpDir(unittest.TestCase):
    '''
    We test this separately since we use it in TestAWSBase.setUp()

    '''
    def setUp(self):
        self.s3_path = 'tmp/' + str(uuid.uuid4()) + '/'

    def tearDown(self):
        b = s3.S3Connection()._bucket(TEST_BUCKET) # pylint: disable=protected-access
        for key in b.list(self.s3_path[0:-1]):
            b.delete_key(key)

    def test_s3_tmpdir(self):
        def fake_uuid():
            fake_uuid.counter += 1
            return "FAKE-UUID-%d" % fake_uuid.counter
        fake_uuid.counter = 0
        self.assertEqual(
            s3.path.gettmpdir(bucket=TEST_BUCKET, prefix=self.s3_path, uuid_generator=fake_uuid),
            's3://%s/%sFAKE-UUID-1/' % (TEST_BUCKET, self.s3_path))
        self.assertEqual(
            len(list(s3.ls('s3://%s/%sFAKE-UUID-1' % (TEST_BUCKET, self.s3_path)))),
            1)
        self.assertTrue(
            s3.exists('s3://%s/%sFAKE-UUID-1/.tempdir' % (TEST_BUCKET, self.s3_path)))
        self.assertEqual(
            s3.path.gettmpdir(bucket=TEST_BUCKET, prefix=self.s3_path, uuid_generator=fake_uuid),
            's3://%s/%sFAKE-UUID-2/' % (TEST_BUCKET, self.s3_path))
        self.assertEqual(len(list(s3.ls('s3://%s/%sFAKE-UUID-2' % (TEST_BUCKET, self.s3_path)))), 1)
        self.assertTrue(s3.exists('s3://%s/%sFAKE-UUID-2/.tempdir' % (TEST_BUCKET, self.s3_path)))



from __future__ import print_function

# Adapted from http://code.activestate.com/recipes/577058/
def confirm(question, default="no"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True or False.
    """
    import sys

    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while 1:
        sys.stdout.write(question + prompt)
        choice = raw_input().lower()
        if default is not None and choice == '':
            return default
        elif choice in valid.keys():
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "\
                             "(or 'y' or 'n').\n")

def create_conditional_print(should_print):
    '''
    When should_print is True, returns a function which prints its inputs.
    When it's false, return a function which does nothing.

    Usage:

    def do_something(verbose=False):
        from bodylabs.util.console import create_conditional_print
        print_verbose = create_conditional_print(verbose)

        print_verbose('Here is something which might print')

    '''
    def noop(*args, **kwargs): # Yup, these args are unused in this no-op function... pylint: disable=unused-argument
        pass

    if should_print:
        return print
    else:
        return noop

def sizeof_format_human_readable(num):
    '''
    Sweet little implementation from
    http://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
    '''
    for x in ['b', 'kb', 'mb', 'gb']:
        if num < 1024.0 and num > -1024.0:
            return "%3.1f%s" % (num, x)
        num /= 1024.0
    return "%3.1f%s" % (num, 'tb')



'''
Usage of parallel_for:

class DoStuffWorker(ParallelWorker):
    def __init__(self, a, b, c):
        # Parameters here can be whatever you want
        self.stuff = (a, b, c)
        super(DoStuffWorker, self).__init__()

    def on_run(self, x, y):
        # Required, executed once for each task.
        # No return value will be collected, expected to have side effects.
        do_stuff(x, y, self.stuff)
        self.output("Woah!") # the output command is intended for things like logging.
                             # It is not one to one with inputs (i.e. this is not a map)

    def on_done(self):
        # Optional, gets excuted after all tasks are done, before the worker dies.
        # Useful for cleanup.
        pass

# Each item in the iterable will be passed to on_run:
lots_o_stuff = ['aa', 'bb', 'cc', 'dd', 'ee']
# If you want have args, make them tuples:
lots_o_stuff = [(1, 1.0), (2, 37.0), (3, 42.0), (4, 3.1415), (5, 91.0)]
output = parallel_for(lots_o_stuff, DoStuffWorker, args=[A, B, C])
'''

import threading

def _parallel_for(q, cls, output, *args, **kwargs):
    worker = cls(*args, **kwargs)
    worker.send_output_to(output)
    worker.run(q)
    output.close()


class MultiPipeWatcher(threading.Thread):
    def __init__(self, pipes):
        super(MultiPipeWatcher, self).__init__()
        self.pipes = pipes
        self.merged = []
        self.running = True

    def run(self):
        while self.running:
            for pipe in self.pipes:
                if pipe.poll():
                    self.merged.append(pipe.recv())
        # flush() has been called. loop over them one more time to empty them out
        for pipe in self.pipes:
            try:
                while pipe.poll():
                    self.merged.append(pipe.recv())
            except EOFError:
                pass
            pipe.close()

    def flush(self):
        self.running = False


def parallel_for(a, cls, args=[], kwargs={}, num_processes=None):
    from multiprocessing import Process, JoinableQueue, cpu_count, Pipe
    if num_processes is None:
        num_processes = cpu_count()
    # Note that JoinableQueue uses an integer for tracking locations in the queue.
    # Because it's using shared memory it's not terribly flexible and gives annoyingly 
    # unclear errors if you go over the limit. We'd like the queue to be as large as
    # possible so that we can avoid contention, but without allocating a max possible
    # size queue unless we need it, thus the calculation below. 32767 is a hard limit.
    q = JoinableQueue(maxsize=min(len(a)+num_processes, 2**15 - 1))

    output_pipes = [Pipe(duplex=False) for _ in range(num_processes)]
    send_pipes = [p for _, p in output_pipes]
    recv_pipes = [p for p, _ in output_pipes]
    pool = [Process(target=_parallel_for, args=(q, cls, pipe) + tuple(args), kwargs=kwargs)
            for pipe in send_pipes]
    output_watcher = MultiPipeWatcher(recv_pipes)
    try:
        for p in pool:
            p.start() 
        output_watcher.start()
        for x in a:
            q.put(x)
        for _ in range(num_processes):
            q.put(None) # End markers
        q.close()
        q.join_thread()
        q.join()
        for p in pool:
            p.join()
        output_watcher.flush()
        output_watcher.join()
        combined_output = output_watcher.merged
        return combined_output
    except KeyboardInterrupt:
        print "Interrupted -- terminating worker processes"
        for p in pool:
            p.terminate() 
        for p in pool:
            p.join() 
        raise


class ParallelWorker(object):
    def __init__(self):
        self.output_pipe = None
        super(ParallelWorker, self).__init__()
    def run(self, q):
        import inspect
        num_args = len(inspect.getargspec(self.on_run).args) - 1 # less self
        while True:
            next_item = q.get()
            if next_item is None:
                # We enque one None per process when we're setting up the queue, so they each find a marker to let them
                # know it's done. Using .empty across processes is just too unstable to be viable.
                q.task_done()
                break
            try:
                if num_args > 1:
                    self.on_run(*next_item)
                else:
                    self.on_run(next_item)
            except Exception: # Really, we want to catch _anything_, otherwise the process will never join pylint: disable=broad-except
                import traceback
                print u"parallel_for FAILED on input {}:".format(next_item)
                traceback.print_exc()
            q.task_done()
        if hasattr(self, 'on_done'):
            self.on_done()

    def send_output_to(self, output):
        self.output_pipe = output

    def output(self, *args):
        self.output_pipe.send(args)



class ProgressCallback(object):
    def __init__(self, progress, maxval=None):
        self.progress = progress
        self.maxval = maxval
        self.minval = 0
        if self.maxval:
            self.progress.maxval = self.maxval
    def __call__(self, done, total=None):
        if total and not self.maxval:
            self.progress.maxval = total
        self.progress.update(done + self.minval)

class BaseProgressbar(object):
    def __init__(self, supress=False, maxval=None):
        self.supress = supress
        self.maxval = maxval
        self.minval = 0
        self.progress = None
    def __enter__(self):
        if not self.supress:
            try:
                self.progress = self.setup_progressbar()
                self.progress.start()
                return ProgressCallback(self.progress, self.maxval)
            except ImportError:
                self.supress = True
                import warnings
                warnings.warn("progressbar not found. Please pip install progressbar")
        if self.supress:
            def cb(*args, **kwargs): # pylint: disable=unused-argument
                pass
            return cb

    def __exit__(self, type, value, traceback): # FIXME pylint: disable=redefined-builtin
        if self.progress and hasattr(self.progress, 'finish'):
            self.progress.finish()

    def setup_progressbar(self):
        from progressbar import ProgressBar, Bar, Percentage
        return ProgressBar(widgets=[Bar(), ' ', Percentage()])

class FileTransferProgressbar(BaseProgressbar):
    def setup_progressbar(self):
        from progressbar import ProgressBar, FileTransferSpeed, Bar, Percentage, ETA
        return ProgressBar(widgets=[FileTransferSpeed(), ' <<<', Bar(), '>>> ', Percentage(), ' ', ETA()])



def md5_for_file(filename, block_size=8192, start=0, end=None):
    '''
    Simple little md5 hash algorithm.
    '''
    import hashlib
    md5 = hashlib.md5()
    with open(filename, 'rb') as f:
        f.seek(start)
        if end:
            while f.tell() + block_size <= end:
                md5.update(f.read(block_size))
            if f.tell() < end:
                md5.update(f.read(end-f.tell()))
        else:
            for chunk in iter(lambda: f.read(block_size), b''):
                md5.update(chunk)
    return md5.hexdigest()



from __future__ import absolute_import

def load(path):
    '''
    from baiji.util import yaml
    foo = yaml.load('foo.yaml')
    '''
    import yaml
    with open(path, 'r') as f:
        return yaml.load(f)



def internet_reachable():
    import socket
    import requests
    try:
        requests.get("http://google.com", timeout=1)
    except (requests.exceptions.Timeout, requests.exceptions.ConnectionError, socket.timeout):
        return False
    return True



import mock
from baiji import s3
from baiji.test_s3 import TestAWSBase

class TestSync(TestAWSBase):
    def setUp(self):
        super(TestSync, self).setUp()

        # Files that are the same on both sides:
        self.create_random_file_at((self.tmp_dir, self.s3_test_location), 'foo/a.txt')
        self.create_random_file_at((self.tmp_dir, self.s3_test_location), 'foo/b.txt')
        self.create_random_file_at((self.tmp_dir, self.s3_test_location), 'foo/bar/c.txt')
        self.create_random_file_at((self.tmp_dir, self.s3_test_location), 'foo/bar/baz/d.txt')
        # Files that are different locally and remotely:
        self.create_random_file_at((self.tmp_dir,), 'foo/x.txt')
        self.create_random_file_at((self.s3_test_location,), 'foo/x.txt')
        self.create_random_file_at((self.tmp_dir,), 'foo/bar/x.txt')
        self.create_random_file_at((self.s3_test_location,), 'foo/bar/x.txt')
        # Files that are only local:
        self.create_random_file_at((self.tmp_dir,), 'foo/loc.txt')
        self.create_random_file_at((self.tmp_dir,), 'foo/bar/loc.txt')
        # Files that are only remote:
        self.create_random_file_at((self.s3_test_location,), 'foo/rem.txt')
        self.create_random_file_at((self.s3_test_location,), 'foo/bar/rem.txt')

        self.local_dir_to_sync = s3.path.join(self.tmp_dir, 'foo')
        self.remote_dir_to_sync = s3.path.join(self.s3_test_location, 'foo')

        self.expected_local_contents = s3.ls(self.local_dir_to_sync)
        self.expected_remote_contents = [x.replace(self.s3_path+'foo/', '') for x in s3.ls(self.remote_dir_to_sync)]

    def create_random_file_at(self, bases, path):
        from baiji.util.testing import random_data
        data = random_data()
        for base in bases:
            with s3.open(s3.path.join(base, path), 'w') as f:
                f.write(data)

    def assertContentsAre(self, expected_contents):
        self.assertSetEqual(set(s3.ls(self.local_dir_to_sync)), set(expected_contents))
        self.assertSetEqual(set([x.replace(self.s3_path+'foo/', '') for x in s3.ls(self.remote_dir_to_sync)]), set(expected_contents))
        # TODO files are equal:


    def test_sync_local_to_remote(self):
        s3.sync(self.local_dir_to_sync, self.remote_dir_to_sync)
        self.assertContentsAre(self.expected_local_contents)

    def test_sync_remote_to_local(self):
        s3.sync(self.remote_dir_to_sync, self.local_dir_to_sync)
        self.assertContentsAre(self.expected_remote_contents)

    @mock.patch('baiji.s3.S3Connection.cp')
    @mock.patch('baiji.s3.S3Connection.rm')
    def test_sync_file_same(self, rm, cp):
        # In these tests, we want to check that rm and cp are invoked only
        # when they should be, so we mock out cp and rm on a new instance of
        # S3Connection. (It seems difficult to use mock.patch on module
        # methods.)
        #
        # We also need s3.path.join but it's defined on the module, not on
        # instances of S3Connection.
        from baiji.s3 import S3Connection

        s3_with_mocks = S3Connection()

        s3_with_mocks.sync_file(
            s3.path.join(self.local_dir_to_sync, 'a.txt'),
            s3.path.join(self.remote_dir_to_sync, 'a.txt')
        )

        self.assertFalse(cp.called)
        self.assertFalse(rm.called)

    @mock.patch('baiji.s3.S3Connection.cp')
    @mock.patch('baiji.s3.S3Connection.rm')
    def test_sync_file_only_src(self, rm, cp):
        from baiji.s3 import S3Connection

        s3_with_mocks = S3Connection()

        s3_with_mocks.sync_file(
            s3.path.join(self.local_dir_to_sync, 'loc.txt'),
            s3.path.join(self.remote_dir_to_sync, 'loc.txt')
        )

        cp.assert_called_once_with(
            s3.path.join(self.local_dir_to_sync, 'loc.txt'),
            s3.path.join(self.remote_dir_to_sync, 'loc.txt'),
            force=False,
            progress=False,
            policy=None,
            encoding=None,
            encrypt=True,
            guess_content_type=False
        )
        self.assertFalse(rm.called)

    @mock.patch('baiji.s3.S3Connection.cp')
    @mock.patch('baiji.s3.S3Connection.rm')
    def test_sync_file_only_dst(self, rm, cp):
        from baiji.s3 import S3Connection

        s3_with_mocks = S3Connection()

        # Test with delete=True and delete=False.

        s3_with_mocks.sync_file(
            s3.path.join(self.local_dir_to_sync, 'rem.txt'),
            s3.path.join(self.remote_dir_to_sync, 'rem.txt'),
            delete=True
        )

        self.assertFalse(cp.called)
        rm.assert_called_once_with(
            s3.path.join(self.remote_dir_to_sync, 'rem.txt')
        )

        rm.reset_mock()

        with self.assertRaises(s3.KeyNotFound):
            s3_with_mocks.sync_file(
                s3.path.join(self.local_dir_to_sync, 'rem.txt'),
                s3.path.join(self.remote_dir_to_sync, 'rem.txt'),
                delete=False
            )

        self.assertFalse(cp.called)
        self.assertFalse(rm.called)

    @mock.patch('baiji.s3.S3Connection.cp')
    @mock.patch('baiji.s3.S3Connection.rm')
    def test_sync_file_exists_but_outdated(self, rm, cp):
        from baiji.s3 import S3Connection

        s3_with_mocks = S3Connection()

        # Test with update=True and update=False.

        s3_with_mocks.sync_file(
            s3.path.join(self.local_dir_to_sync, 'x.txt'),
            s3.path.join(self.remote_dir_to_sync, 'x.txt'),
            update=True
        )

        cp.assert_called_once_with(
            s3.path.join(self.local_dir_to_sync, 'x.txt'),
            s3.path.join(self.remote_dir_to_sync, 'x.txt'),
            force=True,
            progress=False,
            policy=None,
            encoding=None,
            encrypt=True,
            guess_content_type=False
        )
        self.assertFalse(rm.called)

        cp.reset_mock()

        s3_with_mocks.sync_file(
            s3.path.join(self.local_dir_to_sync, 'x.txt'),
            s3.path.join(self.remote_dir_to_sync, 'x.txt'),
            update=False
        )

        self.assertFalse(cp.called)
        self.assertFalse(rm.called)



# https://coderwall.com/p/qawuyq
# Thanks James.

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    long_description = ''
    print 'warning: pandoc or pypandoc does not seem to be installed; using empty long_description'

import os
requirements_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'requirements.txt')
with open(requirements_file, 'r') as f:
    install_requires = [x.strip() for x in f.readlines()]

from setuptools import setup

setup(
    name='baiji',
    version=__import__('baiji').__version__,
    author='Body Labs',
    author_email='alex@bodylabs.com, paul.melnikow@bodylabs.com, chenyang.liu@bodylabs.com',
    description='High-level Python abstraction layer for Amazon S3',
    long_description=long_description,
    url='https://github.com/bodylabs/baiji',
    license='MIT',
    packages=[
        'baiji',
        'baiji/util',
    ],
    scripts=[
        'bin/s3',
    ],
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ]
)



"""
Tools for interacting with Amazon S3.

This module seeks to be higher-level and easier to use than the boto library
provided by Amazon. It abstracts away whether you are writing to S3 or to a
local file, which is important for running code locally, or in an environment
where there is no connection to S3. It also adds some functionality that we
like, such as progress bars, "exclusive create", and context handlers (`with`)
for reading and writing.

The functionality provided by boto is relatively low level: it allows, for
example, setting an S3 key's contents from a string or local file, or pulling
a file from S3 into a local file.

"""
import os
import shutil
import re
from baiji.exceptions import InvalidSchemeException, S3Exception, KeyNotFound, KeyExists, get_transient_error_class
from baiji.util.parallel import ParallelWorker

# FIXME pylint: disable=too-many-lines

S3_MAX_UPLOAD_SIZE = 1024*1024*1024*5 # 5gb

# Path components in our s3 URIs are separated by slashes. Constant included not
# because it should ever change, but for semantic use in code. Remember that os.sep
# is '/' on MacOS and Linux and '\' on Windows.
sep = '/'

# The following convinence functions create a connection to s3 and execute the command

def cp(*args, **kwargs):
    return S3Connection().cp(*args, **kwargs)

def cp_r(*args, **kwargs):
    return S3Connection().cp_r(*args, **kwargs)

def rm(*args, **kwargs):
    return S3Connection().rm(*args, **kwargs)

def rm_r(*args, **kwargs):
    return S3Connection().rm_r(*args, **kwargs)

def ls(*args, **kwargs):
    return S3Connection().ls(*args, **kwargs)

def glob(*args, **kwargs):
    return S3Connection().glob(*args, **kwargs)

def info(*args, **kwargs):
    return S3Connection().info(*args, **kwargs)

def exists(*args, **kwargs):
    return S3Connection().exists(*args, **kwargs)

def size(*args, **kwargs):
    return S3Connection().size(*args, **kwargs)

def etag(*args, **kwargs):
    return S3Connection().etag(*args, **kwargs)

def md5(*args, **kwargs):
    return S3Connection().md5(*args, **kwargs)

def encrypt_at_rest(*args, **kwargs):
    return S3Connection().encrypt_at_rest(*args, **kwargs)

def mv(*args, **kwargs):
    return S3Connection().mv(*args, **kwargs)

def touch(*args, **kwargs):
    return S3Connection().touch(*args, **kwargs)

def sync_file(*args, **kwargs):
    return S3Connection().sync_file(*args, **kwargs)

def sync(*args, **kwargs):
    return S3Connection().sync(*args, **kwargs)

def get_url(*args, **kwargs):
    return S3Connection().get_url(*args, **kwargs)

def put_string(*args, **kwargs):
    return S3Connection().put_string(*args, **kwargs)

def get_string(*args, **kwargs):
    return S3Connection().get_string(*args, **kwargs)

def list_buckets(*args, **kwargs):
    return S3Connection().list_buckets(*args, **kwargs)

def bucket_info(*args, **kwargs):
    return S3Connection().bucket_info(*args, **kwargs)

def create_bucket(*args, **kwargs):
    return S3Connection().create_bucket(*args, **kwargs)

def open(key, mode='rb'): # pylint: disable=redefined-builtin
    '''
    Acts like open(key, mode), opening a file.

    If the file is on S3, it is downloaded and a NamedTemporaryFile is returned.

    The caller is responsible for closing the file descriptor, please try to
    use it in a with block.

    Like os.open, after stripping 'U' from the mode, the first character must
    be 'r' or 'w'. open also accepts 'x' for exclusive creation: it raises an
    exception if the file already exists. This is meant as a convenience and
    sanity check when you know the original file isn't supposed to exist.

    Note that 'x' is not suitable for preventing concurrent creation. This is
    due to the "eventually consistent" nature of s3, and also the CachedFile
    design. If two processes simultaneously write to the same file, there's a
    race condition. One or both will appear to succeed, and one will win out.
    Using 'x' doesn't -- can't -- prevent this.

    Raises s3.KeyNotFound when attempting to open a local or remote file that
    doesn't exist, s3.KeyExists when attempting exclusive create on an
    existing file or key, ValueError for an invalid key, and IOError for an
    underlying local file-system failure.

    '''
    return CachedFile(key, mode)


def _strip_initial_slashes(key):
    '''
    Keys going into boto need to not have initial slashes.
    '''
    return re.sub(r'^/*', '', key)


class S3CopyOperation(object):
    class CopyableKey(object):
        def __init__(self, key, connection):
            self.raw = key
            self.connection = connection
            self.parsed = path.parse(key)
            self.remote_path = None # value here will be set by the path setting, this just satisfies lint
            self.path = self.parsed.path
            if not (self.path.startswith(sep) or re.match(r'^[a-zA-Z]:', self.path)):
                self.path = sep + self.path
            self.bucket_name = self.parsed.netloc
            self.scheme = self.parsed.scheme
            if self.scheme not in ['file', 's3']:
                raise InvalidSchemeException("URI Scheme %s is not implemented" % self.scheme)
        @property
        def path(self):
            return self._path
        @path.setter
        def path(self, val):
            self._path = val # we get initialized with a call to the setter in init pylint: disable=attribute-defined-outside-init
            # For remote operations, we need a path without initial slashes
            self.remote_path = _strip_initial_slashes(self.path)
        @property
        def bucket(self):
            return self.connection._bucket(self.bucket_name) # pylint: disable=protected-access
        @property
        def uri(self):
            if self.is_file:
                return self.path
            else:
                return "s3://" + self.bucket_name + self.path
        @property
        def is_file(self):
            return self.scheme == 'file'
        @property
        def is_s3(self):
            return self.scheme == 's3'
        def exists(self):
            return self.connection.exists(self.uri)
        def etag(self):
            return self.connection.etag(self.uri)
        def rm(self):
            return self.connection.rm(self.uri)
        def lookup(self):
            if self.is_file:
                raise ValueError("S3CopyOperation.CopyableKey.lookup called for local file")
            key = self.bucket.lookup(self.remote_path)
            if not key:
                raise KeyNotFound("Error finding %s on s3: doesn't exist" % (self.uri))
            return key
        def create(self):
            if self.is_file:
                raise ValueError("S3CopyOperation.CopyableKey.create called for local file")
            from boto.s3.key import Key
            key = Key(self.bucket)
            key.key = self.remote_path
            return key

    def __init__(self, src, dst, connection):
        '''
        Both src and dst may be files or s3 keys
        '''
        self.connection = connection
        self.src = self.CopyableKey(src, connection)
        self.dst = self.CopyableKey(dst, connection)
        self.task = (self.src.scheme, self.dst.scheme)

        if path.isdirlike(self.dst.path):
            self.dst.path = os.path.join(self.dst.path, os.path.basename(self.src.path))

        # DEFAULTS:
        self.progress = False
        self.force = False
        self.policy = None
        self.preserve_acl = False
        self.encrypt = False
        self.encoding = None
        self.gzip = False
        self.content_type = None
        self.metadata = {}
        self.skip = False

        self.retries_allowed = 1
        self._retries = 0

        self.file_size = None

    @property # read only
    def retries_made(self):
        return self._retries

    @property
    def policy(self):
        return self._policy
    @policy.setter
    def policy(self, val):
        if val and self.dst.is_file:
            raise ValueError("Policy only allowed when copying to s3")
        self._policy = val  # we get initialized with a call to the setter in init pylint: disable=attribute-defined-outside-init

    @property
    def preserve_acl(self):
        return self._preserve_acl
    @preserve_acl.setter
    def preserve_acl(self, val):
        val = bool(val)
        if val and self.task != ('s3', 's3'):
            raise ValueError("Preserve ACL only allowed when copying from s3 to s3")
        self._preserve_acl = val  # we get initialized with a call to the setter in init pylint: disable=attribute-defined-outside-init

    @property
    def progress(self):
        return self._progress
    @progress.setter
    def progress(self, val):
        val = bool(val)
        self._progress = val  # we get initialized with a call to the setter in init pylint: disable=attribute-defined-outside-init

    @property
    def force(self):
        return self._force
    @force.setter
    def force(self, val):
        val = bool(val)
        self._force = val  # we get initialized with a call to the setter in init pylint: disable=attribute-defined-outside-init

    @property
    def skip(self):
        return self._skip_exist
    @skip.setter
    def skip(self, val):
        val = bool(val)
        self._skip_exist = val # we get initialized with a call to the setter in init pylint: disable=attribute-defined-outside-init

    @property
    def encrypt(self):
        return self._encrypt
    @encrypt.setter
    def encrypt(self, val):
        val = bool(val) and self.dst.is_s3
        self._encrypt = val  # we get initialized with a call to the setter in init pylint: disable=attribute-defined-outside-init

    @property
    def validate(self):
        return self._validate
    @validate.setter
    def validate(self, val):
        self._validate = bool(val) # we get initialized with a call to the setter in init pylint: disable=attribute-defined-outside-init

    @property
    def encoding(self):
        return self._encoding
    @encoding.setter
    def encoding(self, val):
        if val is not None and self.dst.is_file:
            raise ValueError("Encoding can only be specified when copying to s3")
        if val is not None and self.gzip and val != 'gzip':
            raise ValueError("gzip overrides explicit encoding")
        self._encoding = val # we get initialized with a call to the setter in init pylint: disable=attribute-defined-outside-init

    @property
    def gzip(self):
        return self._gzip
    @gzip.setter
    def gzip(self, val):
        val = bool(val)
        if val and self.task != ('file', 's3'):
            raise ValueError("gzip can only be specified when uploading to s3")
        if val and self.encoding is not None and self.encoding != 'gzip':
            raise ValueError("gzip overrides explicit encoding")
        self._gzip = val # we get initialized with a call to the setter in init pylint: disable=attribute-defined-outside-init

    @property
    def content_type(self):
        return self._content_type
    @content_type.setter
    def content_type(self, val):
        if val is not None and self.dst.is_file:
            raise ValueError("Content Type can only be specified when copying to s3")
        self._content_type = val # we get initialized with a call to the setter in init pylint: disable=attribute-defined-outside-init
    def guess_content_type(self):
        import mimetypes
        content_type, _ = mimetypes.guess_type(self.src.path)
        if not content_type and self.is_extensionless_html():
            content_type = 'text/html'
        elif not content_type:
            content_type = 'application/octet-stream'
        self.content_type = content_type
    def is_extensionless_html(self):
        # test if file has no extension and html5 doctype in first 100 bytes
        if os.path.splitext(self.src.path)[-1] == '':
            if self.encoding == 'gzip':
                import gzip
                with gzip.open(self.src.path, 'rb') as f:
                    return '<!DOCTYPE html>' in f.read(100)
            else:
                with open(self.src.path, 'r') as f:
                    return '<!DOCTYPE html>' in f.read(100)
        else:
            return False


    @property
    def metadata(self):
        meta = self._metadata
        if self.encoding is not None:
            meta['Content-Encoding'] = self.encoding
        if self.content_type is not None:
            meta['Content-Type'] = self.content_type
        return meta
    @metadata.setter
    def metadata(self, val):
        if not (val is None or val == {}) and self.dst.is_file:
            raise ValueError("Metadata can only be specified when copying to s3")
        if val is None:
            val = {}
        self._metadata = val # we get initialized with a call to the setter in init pylint: disable=attribute-defined-outside-init

    def execute(self):
        from boto.s3.connection import S3ResponseError
        if not self.force and self.dst.exists():
            if self.skip:
                import warnings
                warnings.warn("Skipping existing destination copying %s to %s: Destinaton exists" % (self.src.uri, self.dst.uri))
                return
            else:
                raise KeyExists("Error copying %s to %s: Destinaton exists" % (self.src.uri, self.dst.uri))

        if self.dst.is_file:
            self.prep_local_destination()

        try:
            if self.task == ('file', 'file'):
                self.local_copy()
            elif self.task == ('file', 's3'):
                self.upload()
            elif self.task == ('s3', 'file'):
                self.download()
            elif self.task == ('s3', 's3'):
                self.remote_copy()
            else:
                raise InvalidSchemeException("Copy for URI Scheme %s to %s is not implemented" % self.task)

        except (IOError, KeyNotFound):
            raise KeyNotFound("Error copying %s to %s: Source doesn't exist" % (self.src.uri, self.dst.uri))
        except S3ResponseError as e:
            if e.status == 403:
                raise S3Exception("HTTP Error 403: Permission Denied on {}".format(" or ".join([x.uri for x in [self.src, self.dst] if x.is_s3])))
            else:
                raise

    def local_copy(self):
        shutil.copy(self.src.path, self.dst.path)

    def upload(self):
        if self.gzip:
            import gzip
            from baiji.util import tempfile
            with tempfile.NamedTemporaryFile() as compressed_src:
                with open(self.src.path, 'rb') as f:
                    with gzip.open(compressed_src.name, 'wb') as tf:
                        tf.writelines(f)
                self.src.path = compressed_src.name
                self.encoding = 'gzip'
                self._upload()
        else:
            self._upload()

    def _upload(self):
        self.file_size = os.path.getsize(self.src.path)
        if self.file_size > S3_MAX_UPLOAD_SIZE:
            self.upload_multipart()
        else:
            self.upload_direct()

    def upload_multipart(self):
        import math
        from baiji.util.with_progressbar import FileTransferProgressbar
        with FileTransferProgressbar(supress=(not self.progress), maxval=self.file_size) as cb:
            mp = self.dst.bucket.initiate_multipart_upload(self.dst.remote_path, encrypt_key=self.encrypt, metadata=self.metadata)
            n_parts = int(math.ceil(float(self.file_size) / S3_MAX_UPLOAD_SIZE))
            try:
                for ii in range(n_parts):
                    with open(self.src.path, 'r') as fp:
                        fp.seek(S3_MAX_UPLOAD_SIZE*ii)
                        part_size = self.file_size - S3_MAX_UPLOAD_SIZE*ii if ii+1 == n_parts else S3_MAX_UPLOAD_SIZE
                        cb.minval = S3_MAX_UPLOAD_SIZE*ii
                        num_cb = part_size / (1024*1024)
                        mp.upload_part_from_file(fp=fp, part_num=ii+1, size=part_size, cb=cb, num_cb=num_cb)
                mp.complete_upload()
            except:
                mp.cancel_upload()
                raise

    def upload_direct(self):
        import math
        from baiji.util.with_progressbar import FileTransferProgressbar
        key = self.dst.create()
        for k, v in self.metadata.items():
            key.set_metadata(k, v)
        with FileTransferProgressbar(supress=(not self.progress)) as cb:
            num_cb = max([10, int(math.ceil(float(self.file_size) / (1024*1024)))])
            key.set_contents_from_filename(self.src.path, cb=cb, policy=self.policy, encrypt_key=self.encrypt, num_cb=num_cb)

    def download(self):
        '''
        Download to local file

        If `validate` is set, check etags and retry once if not match.
        Raise TransientError when download is corrupted again after retry

        '''
        from baiji.util import tempfile
        from baiji.util.with_progressbar import FileTransferProgressbar
        # We create, close, and delete explicitly rather than using
        # a `with` block, since on windows we can't have a file open
        # twice by the same process.
        tf = tempfile.NamedTemporaryFile(delete=False)
        try:
            key = self.src.lookup()

            with FileTransferProgressbar(supress=(not self.progress)) as cb:
                key.get_contents_to_file(tf, cb=cb)
            tf.close()

            if self.validate:
                self.ensure_integrity(tf.name)

            # We only actually write to dst.path if the download succeeds and
            # if necessary is validated. This avoids leaving partially
            # downloaded files if something goes wrong.
            shutil.copy(tf.name, self.dst.path)

        except get_transient_error_class() as transient_error:
            # Printed here so that papertrail can alert us when this occurs
            print transient_error

            # retry once or raise
            if self.retries_made < self.retries_allowed:
                self._retries += 1
                self.download()
            else:
                raise

        finally:
            self.connection.rm(tf.name)

    def ensure_integrity(self, filename):
        '''
        Ensure integrity of downloaded file; raise TransientError if there's a mismatch
        '''
        if self.connection.etag(filename) != self.src.etag():
            raise get_transient_error_class()('Destinaton file for ({}) is corrupted, retry count {}'.format(self.src.uri, self.retries_made))

    def remote_copy(self):
        '''
        With copy_key, if metadata is None, it will be copied from the existing key
        '''
        src = self.src.remote_path
        headers = {}
        if self.policy:
            headers['x-amz-acl'] = self.policy
        key = self.src.lookup()
        meta = key.metadata
        meta['Content-Encoding'] = key.content_encoding
        meta['Content-Type'] = key.content_type
        meta = dict(meta.items() + self.metadata.items())
        self.dst.bucket.copy_key(self.dst.remote_path, self.src.bucket_name, src, preserve_acl=self.preserve_acl, metadata=meta, headers=headers, encrypt_key=self.encrypt)
        if self.progress:
            print 'Copied %s to %s' % (self.src.uri, self.dst.uri)

    def prep_local_destination(self):
        from baiji.util.shutillib import mkdir_p
        mkdir_p(os.path.dirname(self.dst.path))


class MultifileCopyWorker(ParallelWorker):
    '''
    This is a pickleable worker used for parallel copy operations.
    '''
    def __init__(self, kwargs_for_cp):
        self.connection = S3Connection(cache_buckets=True)
        self.kwargs_for_cp = kwargs_for_cp
        self.verbose = self.kwargs_for_cp.get('progress', False)
        if 'progress' in self.kwargs_for_cp:
            self.kwargs_for_cp['progress'] = False
        super(MultifileCopyWorker, self).__init__()

    def on_run(self, file_from, file_to):
        try:
            self.connection.cp(file_from, file_to, **self.kwargs_for_cp)
            if self.verbose:
                print "Finished transfering {} to {}".format(file_from, file_to)
        except KeyExists as e:
            # Note here that the correct behavior would probably be to roll back
            # if we encounter an error, but that's not practical, so we copy
            # what we can and show an error about the rest.
            print str(e)


class S3Connection(object):
    def __init__(self, cache_buckets=False):
        self._connected = False
        self._conn = None
        self.cache_buckets = cache_buckets
        self._known_valid_buckets = set()

    @property
    def conn(self):
        from boto.s3.connection import S3Connection as BotoS3Connection
        from boto.s3.connection import OrdinaryCallingFormat
        from baiji.config import credentials
        if not self._connected:
            self._conn = BotoS3Connection(credentials.key, credentials.secret,
                                          calling_format=OrdinaryCallingFormat(),
                                          suppress_consec_slashes=False)
            self._connected = True
        return self._conn

    def _bucket(self, name, cache_buckets=None):
        '''
        The call to `get_bucket` will make a HEAD request to S3 and raise S3ResponseError if the bucket doesn't exist.
        If we pass `validate=False` then it won't make the request, but it's behavor will be undefined if the bucket doesn't
        exist. So we don't want to not validate buckets in general, because that'd break things, but if we've seen the bucket
        before, then it's very likely that we'll see it again.

        We support an object wide cache_buckets, or we can pass cache_buckets just to this call, which will override the
        object wide value.
        '''
        from boto.s3.connection import S3ResponseError
        if cache_buckets is None:
            cache_buckets = self.cache_buckets
        try:
            if cache_buckets:
                if name in self._known_valid_buckets:
                    return self.conn.get_bucket(name, validate=False)
                else:
                    bucket = self.conn.get_bucket(name) # this will raise if the bucket isn't valid
                    self._known_valid_buckets.add(name)
                    return bucket
            else:
                return self.conn.get_bucket(name)
        except S3ResponseError as e:
            if e.status == 403:
                raise S3Exception("HTTP Error 403: Permission Denied on s3://{}/".format(name))
            else:
                raise

    def _lookup(self, bucket_name, key, cache_buckets=None):
        '''
        See _bucket for the details on cache_buckets
        '''
        key = _strip_initial_slashes(key)
        return self._bucket(bucket_name, cache_buckets=cache_buckets).lookup(key)

    def cp(self, key_or_file_from, key_or_file_to, force=False, progress=False, policy=None, preserve_acl=False, encoding=None, encrypt=True, gzip=False, content_type=None, guess_content_type=False, metadata=None, skip=False, validate=True):
        """
        Copy file to or from AWS S3

        Both the from and to arguments can be either local paths or s3 urls (e.g. ``s3://bucketname/path/in/bucket``).

        If the target exists, the an exception is raised unless ``force=True`` is given.

        When copying from file to s3, you can optionally specify a canned acl policy,
        e.g. `policy='public-read'`. For possible values, see boto.s3.acl.CannedACLStrings.

        When copying from s3 to s3, you can specify preserve_acl=True to copy the
        existing ACL to the new file. With preserve_acl=False it will inherit the
        new bucket's policy.
        """

        op = S3CopyOperation(key_or_file_from, key_or_file_to, connection=self)
        op.progress = progress
        op.force = force
        op.policy = policy
        op.preserve_acl = preserve_acl
        op.encrypt = encrypt
        op.encoding = encoding
        op.gzip = gzip
        op.metadata = metadata
        op.skip = skip
        op.validate = validate

        if guess_content_type:
            op.guess_content_type()
        else:
            op.content_type = content_type

        op.execute()

    def cp_r(self, dir_from, dir_to, parallel=False, **kwargs):
        '''
        kwargs are passed on directly to s3.cp; see defaults there.
        '''
        (from_scheme, _, from_path, _, _, _) = path.parse(dir_from)
        if from_scheme == 'file':
            files_to_copy = [(path.join(dir_from, f), path.join(dir_to, f))
                             for f in ls(dir_from, return_full_urls=False)]
        else:
            if from_path.endswith(sep):
                # Emulate `cp`, which copies the contents of the path.
                # Get path relative to from_path
                files_to_copy = [(f, path.join(dir_to, os.path.relpath(path.parse(f).path, from_path)))
                                 for f in ls(dir_from, return_full_urls=True)]
            else:
                # Get path relative to from_path's parent
                # Since from_path has no '/', we can get this with os.path.dirname()
                files_to_copy = [(f, path.join(dir_to, os.path.relpath(path.parse(f).path, os.path.dirname(from_path))))
                                 for f in ls(dir_from, return_full_urls=True)]

        if 'force' not in kwargs or not kwargs['force']:
            # we're not supposed to overwrite. Locally this is easy, since `exists` checks are cheap, but
            # on s3, it's more expensive, so we avoid it if possible:
            if path.isremote(dir_to):
                def common_prefix(a, b):
                    try:
                        ind = [x == y for x, y in zip(a, b)].index(False)
                    except ValueError:
                        return a
                    return a[:ind]
                destinations = [y for _, y in files_to_copy]
                prefix = reduce(common_prefix, destinations[1:], destinations[0])
                try:
                    # note that we can't use `exists` here, as it only works for full keys
                    self.ls(prefix).next()
                except StopIteration:
                    # There's nothing in the iterator, so there are no files to be found, so
                    # we set force for the copy so that we don't have to check each one:
                    kwargs['force'] = True

        if parallel:
            from baiji.util.parallel import parallel_for
            parallel_for(files_to_copy, MultifileCopyWorker, args=[kwargs], num_processes=12)
        else:
            for file_from, file_to in files_to_copy:
                try:
                    # Note here that the correct behavior would probably be to roll back
                    # if we encounter an error, but that's not practical, so we copy
                    # what we can and show an error about the rest.
                    self.cp(file_from, file_to, **kwargs)
                except KeyExists as e:
                    print str(e)

    def rm(self, key_or_file):
        '''
        Remove a key from AWS S3
        '''
        k = path.parse(key_or_file)
        if k.scheme == 'file':
            if os.path.isdir(k.path):
                shutil.rmtree(k.path)
            elif os.path.exists(k.path):
                return os.remove(k.path)
            else:
                raise KeyNotFound("%s does not exist" % key_or_file)
        elif k.scheme == 's3':
            if not exists(key_or_file):
                raise KeyNotFound("%s does not exist" % key_or_file)
            return self._bucket(k.netloc).delete_key(_strip_initial_slashes(k.path))
        else:
            raise InvalidSchemeException("URI Scheme %s is not implemented" % k.scheme)

    def rm_r(self, key_or_file, force=False, quiet=False):
        '''
        Prompts for confirmation on each file when force is False.

        Raises an exception when not using AWS.
        '''
        k = path.parse(key_or_file)
        if not k.scheme == 's3':
            raise InvalidSchemeException("URI Scheme %s is not implemented" % k.scheme)
        bucket = k.netloc
        keys_to_delete = self.ls(key_or_file)
        for key_to_delete in keys_to_delete:
            url = "s3://%s%s" % (bucket, key_to_delete)
            if not force:
                from baiji.util.console import confirm
                if not confirm("Remove %s" % url):
                    continue
            self.rm(url)
            if not quiet:
                print "[deleted] %s" % url

    def ls(self, s3prefix, return_full_urls=False, require_s3_scheme=False, shallow=False, followlinks=False):
        '''
        List files on AWS S3
        prefix is given as an S3 url: ``s3://bucket-name/path/to/dir``.
        It will return all values in the bucket that have that prefix.

        Note that ``/dir/filename.ext`` is found by ``ls('s3://bucket-name/dir/fil')``; it's really a prefix and not a directory name.

        A local prefix generally is acceptable, but if require_s3_scheme
        is True, the prefix must be an s3 URL.

        If `shallow` is `True`, the key names are processed hierarchically
        using '/' as a delimiter, and only the immediate "children" are
        returned.

        '''
        import itertools
        k = path.parse(s3prefix)
        if k.scheme == 's3':
            prefix = k.path
            if prefix.startswith(sep):
                prefix = prefix[len(sep):]
            delimiter = shallow and sep or ''
            if return_full_urls:
                clean_paths = lambda x: "s3://" + k.netloc + sep + x.name
            else:
                clean_paths = lambda x: sep + x.name
            return itertools.imap(clean_paths, self._bucket(k.netloc).list(prefix=prefix, delimiter=delimiter))
        elif k.scheme == 'file':
            if require_s3_scheme:
                raise InvalidSchemeException('URI should begin with s3://')
            paths = []
            remove = ''
            if not return_full_urls:
                remove = k.path
                if not remove.endswith(os.sep):
                    remove += os.sep
            for root, _, files in os.walk(k.path, followlinks=followlinks):
                for f in files:
                    # On Windows, results of os.abspath() and os.walk() have '\',
                    # so we replace them with '/'
                    paths.append(path.join(root, f).replace(remove, '').replace(os.sep, sep))
                if shallow:
                    break
            return paths
        else:
            raise InvalidSchemeException("URI Scheme %s is not implemented" % k.scheme)

    def glob(self, prefix, pattern):
        '''
        Given a path prefix and a pattern, iterate over matching paths.

        e.g.

        paths = list(s3.glob(
            prefix='s3://bodylabs-ants-go-marching/output/feet_on_floor/eff2a0e/',
            pattern='*_alignment.ply'
        ))

        '''
        import fnmatch
        import functools
        import itertools
        predicate = functools.partial(fnmatch.fnmatch, pat=prefix + pattern)
        listing = ls(prefix, return_full_urls=True)
        return itertools.ifilter(predicate, listing)

    def info(self, key_or_file):
        '''
        Get info about a file
        '''
        from datetime import datetime
        k = path.parse(key_or_file)
        result = {
            'uri': '%s://%s%s' % (k.scheme, k.netloc, k.path),
        }
        if k.scheme == 'file':
            if not os.path.exists(k.path):
                raise KeyNotFound("Error getting info on %s: File doesn't exist" % (key_or_file, ))
            stat = os.stat(k.path)
            result['size'] = stat.st_size
            result['last_modified'] = datetime.fromtimestamp(stat.st_mtime)
        elif k.scheme == 's3':
            remote_object = self._lookup(k.netloc, k.path)
            if remote_object is None:
                raise KeyNotFound("Error getting info on %s: Key doesn't exist" % (key_or_file, ))
            result['size'] = remote_object.size
            result['last_modified'] = datetime.strptime(remote_object.last_modified, "%a, %d %b %Y %H:%M:%S GMT")
            result['content_type'] = remote_object.content_type
            result['content_encoding'] = remote_object.content_encoding
            result['encrypted'] = bool(remote_object.encrypted)
            result['acl'] = remote_object.get_acl()
            result['owner'] = remote_object.owner
        else:
            raise InvalidSchemeException("URI Scheme %s is not implemented" % k.scheme)
        return result

    def exists(self, key_or_file, retries_allowed=3):
        '''
        Check if a file exists on AWS S3

        Returns a boolean.

        If the key is not found then we recheck up to `retries_allowed` times. We only do this
        on s3. We've had some observations of what appears to be eventual consistency, so this
        makes it a bit more reliable. This does slow down the call in the case where the key
        does not exist. 

        On a relatively slow, high latency connection a test of 100 tests retreiving a
        non-existant file gives:

        With retries_allowed=1: median=457.587 ms, mean=707.12387 ms
        With retries_allowed=3: median=722.969 ms, mean=1185.86299 ms
        with retries_allowed=10: median=2489.767 ms, mean=2995.34233 ms
        With retries_allowed=100: median=24694.0815 ms, mean=26754.64137 ms

        So assume that letting retries_allowed=3 will cost you a bit less than double the time.
        '''
        k = path.parse(key_or_file)
        if k.scheme == 'file':
            return os.path.exists(k.path)
        elif k.scheme == 's3':
            retry_attempts = 0
            while retry_attempts < retries_allowed:
                key_exists = self._lookup(k.netloc, k.path, cache_buckets=True) is not None
                if key_exists:
                    if retry_attempts > 0: # only if we find it after failing at least once
                        import warnings
                        from baiji.exceptions import EventualConsistencyWarning
                        warnings.warn("S3 is behaving in an eventually consistent way in s3.exists({}) -- it took {} attempts to locate the key".format(key_or_file, retry_attempts+1), EventualConsistencyWarning)
                    return True
                retry_attempts += 1
            return False
        else:
            raise InvalidSchemeException("URI Scheme %s is not implemented" % k.scheme)

    def size(self, key_or_file):
        '''
        Return the size of a file. If it's on s3, don't download it.
        '''
        k = path.parse(key_or_file)
        if k.scheme == 'file':
            return os.path.getsize(k.path)
        elif k.scheme == 's3':
            k = self._lookup(k.netloc, k.path)
            if k is None:
                raise KeyNotFound("s3://%s/%s not found on s3" % (k.netloc, k.path))
            return k.size
        else:
            raise InvalidSchemeException("URI Scheme %s is not implemented" % k.scheme)

    def _get_etag(self, netloc, remote_path):
        k = self._lookup(netloc, remote_path)
        if k is None:
            raise KeyNotFound("s3://%s/%s not found on s3" % (netloc, remote_path))
        return k.etag.strip("\"") # because s3 seriously gives the md5sum back wrapped in an extra set of double quotes...

    def etag(self, key_or_file):
        '''
        Return the s3 etag of the file. For files less than 5gb, this is the same as md5.
        '''
        k = path.parse(key_or_file)
        if k.scheme == 'file':
            import math
            import struct
            import hashlib
            from baiji.util.md5 import md5_for_file
            file_size = os.path.getsize(k.path)
            if file_size > S3_MAX_UPLOAD_SIZE:
                n_parts = int(math.ceil(float(file_size) / S3_MAX_UPLOAD_SIZE))
                starts = [ii*S3_MAX_UPLOAD_SIZE for ii in range(n_parts)]
                ends = [(ii+1)*S3_MAX_UPLOAD_SIZE for ii in range(n_parts)]
                ends[-1] = None
                hashes = [md5_for_file(k.path, start=start, end=end) for start, end in zip(starts, ends)]
                md5_digester = hashlib.md5()
                for h in hashes:
                    md5_digester.update(struct.pack("!16B", *[int(h[x:x+2], 16) for x in range(0, len(h), 2)]))
                return md5_digester.hexdigest() + "-%d" % n_parts
            else:
                return md5_for_file(k.path)
        elif k.scheme == 's3':
            return self._get_etag(k.netloc, k.path)
        else:
            raise InvalidSchemeException("URI Scheme %s is not implemented" % k.scheme)

    def md5(self, key_or_file):
        '''
        Return the MD5 checksum of a file. If it's on s3, don't download it.
        '''
        k = path.parse(key_or_file)
        if k.scheme == 'file':
            from baiji.util.md5 import md5_for_file
            return md5_for_file(k.path)
        elif k.scheme == 's3':
            res = self._get_etag(k.netloc, k.path)
            if "-" in res:
                raise Exception("md5 hashes not availiable from s3 for files over 5gb")
            return res
        else:
            raise InvalidSchemeException("URI Scheme %s is not implemented" % k.scheme)

    def encrypt_at_rest(self, key):
        '''
        This method takes a key on s3 and encrypts it.
        Note that calling this method on a local file is an error
        and that calling it on an s3 key that is already encrypted,
        while allowed, is a no-op.
        '''
        k = path.parse(key)
        if k.scheme != 's3':
            raise InvalidSchemeException("URI Scheme %s is not implemented" % k.scheme)
        remote_object = self._lookup(k.netloc, k.path)
        if remote_object is None:
            raise KeyNotFound("Error encrypting %s: Key doesn't exist" % (key, ))
        if not bool(remote_object.encrypted):
            bucket = self._bucket(k.netloc)
            src = k.path
            if src.startswith(sep):
                src = src[len(sep):] # NB: copy_key is failing with absolute src keys...
            bucket.copy_key(src, k.netloc, src, preserve_acl=True, metadata=None, encrypt_key=True)

    def mv(self, key_or_file_from, key_or_file_to, **kwargs):
        """
        Move file to or from AWS S3

        Both the from and to arguments can be either local paths or s3 urls (e.g. ``s3://bucketname/path/in/bucket``).

        If the target exists, the an exception is raised unless ``force=True`` is given.
        kwargs are passed directly on to s3.cp; see there for defaults.
        """
        self.cp(key_or_file_from, key_or_file_to, **kwargs)
        self.rm(key_or_file_from)

    def touch(self, key, encrypt=True):
        """
        Touch a local file or a path on s3
        
        Locally, this is analagous to the unix touch command
        
        On s3, it creates an empty file if there is not one there already,
        but does not change the timestamps (not possible to do without
        actually moving the file)
        """
        if path.islocal(key):
            from bodylabs.util.shutillib import touch as _touch
            _touch(path.parse(key).path)
        else:
            # The replace=False here means that we only take action if
            # the file doesn't exist, so we don't accidentally truncate
            # files when we just mean to be touching them
            self.put_string(key, '', encrypt=encrypt, replace=False)

    def sync_file(self, src, dst, update=True, delete=False, progress=False, policy=None, encoding=None, encrypt=True, guess_content_type=False):
        '''
        Sync a file from src to dst.

        update: When True, update dst if it exists but contents do not match.
        delete: When True, remove dst if src does not exist. When False, raise
          an error if src does not exist.
        '''
        from baiji.util.console import create_conditional_print
        print_verbose = create_conditional_print(progress)

        src_exists = self.exists(src)
        if not delete and not src_exists:
            raise KeyNotFound(
                "Error syncing {} to {}: Source doesn't exist".format(src, dst))

        dst_exists = self.exists(dst)

        needs_delete = dst_exists and not src_exists
        needs_fresh_copy = src_exists and not dst_exists
        needs_update = dst_exists and src_exists and self.etag(src) != self.etag(dst)

        if not needs_delete and not needs_fresh_copy and not needs_update:
            print_verbose('{} is up to date'.format(dst))
            return

        # At this point, exactly one of these should be true.
        assert needs_delete ^ needs_fresh_copy ^ needs_update

        if needs_fresh_copy:
            print_verbose('copying {} to {}'.format(src, dst))
            self.cp(src, dst, force=False, progress=progress, policy=policy, encoding=encoding, encrypt=encrypt, guess_content_type=guess_content_type)
        elif needs_update:
            print_verbose('file is out of date: {}'.format(dst))
            if update:
                print_verbose('copying {} to {}'.format(src, dst))
                self.cp(src, dst, force=True, progress=progress, policy=policy, encoding=encoding, encrypt=encrypt, guess_content_type=guess_content_type)
        elif needs_delete:
            print_verbose('source file does not exist: {}'.format(src))
            if delete:
                print_verbose('removing {}'.format(dst))
                self.rm(dst)

    def sync(self, key_or_dir_from, key_or_dir_to, followlinks=True, do_not_delete=None, progress=False, policy=None, encoding=None, encrypt=True, guess_content_type=False):
        '''
        Sync a directory of files.

        Note these src and dst paths must be directories. Use `s3.sync_file`
        for files.

        do_not_delete: Paths not to delete during syncing, useful for syncing root
          content in a bucket when other apps are responsible for filling in
          subfolders.

        '''
        def clean_path(key_or_dir):
            if path.islocal(key_or_dir) and not os.path.isdir(key_or_dir):
                raise ValueError("s3 sync requires directories")
            if path.islocal(key_or_dir):
                key_or_dir = os.path.abspath(key_or_dir)
            if not key_or_dir.endswith('/'):
                key_or_dir += '/'
            return key_or_dir

        key_or_dir_from = clean_path(key_or_dir_from)
        key_or_dir_to = clean_path(key_or_dir_to)

        src_files = [x.replace(key_or_dir_from, '') for x in self.ls(key_or_dir_from, return_full_urls=True, followlinks=followlinks)]
        dst_files = [x.replace(key_or_dir_to, '') for x in self.ls(key_or_dir_to, return_full_urls=True, followlinks=followlinks)]
        for f in src_files:
            src = key_or_dir_from + f
            dst = key_or_dir_to + f
            self.sync_file(src, dst, progress=progress, policy=policy, encoding=encoding, encrypt=encrypt, guess_content_type=guess_content_type)

        do_not_delete = [] if do_not_delete is None else do_not_delete
        do_not_delete = [x if x.endswith('/') else x + '/' for x in do_not_delete]
        for f in set(dst_files) - set(src_files):
            removed_file = key_or_dir_to + f
            if any([f.startswith(x) for x in do_not_delete]):
                if progress:
                    print "leaving alone", removed_file
            else:
                if progress:
                    print "removing", removed_file
                self.rm(removed_file)

    def get_url(self, key, ttl):
        """
        Get a temporary https url for a file on AWS S3

        Returns the url as a string.
        The url will timeout and return an error after ``ttl`` seconds.
        """
        k = path.parse(key)
        return self._lookup(k.netloc, k.path).generate_url(ttl)

    def put_string(self, key, s, encrypt=True, replace=True):
        '''
        Save string ``s`` to S3 as ``key``.

        If ``replace=True``, this will overwrite an existing key.
        If ``replace=false``, this will be a no-op when the key already exists.

        '''
        from boto.s3.key import Key
        key = path.parse(key)
        b = self._bucket(key.netloc)
        k = Key(b)
        k.key = _strip_initial_slashes(key.path)
        k.set_contents_from_string(s, encrypt_key=encrypt, replace=replace)

    def get_string(self, key):
        '''
        Get string stored in S3 ``key``.
        '''
        k = path.parse(key)
        return self._lookup(k.netloc, k.path).get_contents_as_string()

    def open(self, key, mode='rb'): # pylint: disable=redefined-builtin
        '''
        Acts like open(key, mode), opening a file.

        If the file is on S3, it is downloaded and a NamedTemporaryFile is returned.

        The caller is responsible for closing the file descriptor, please try to
        use it in a with block.

        Like os.open, after stripping 'U' from the mode, the first character must
        be 'r' or 'w'. open also accepts 'x' for exclusive creation: it raises an
        exception if the file already exists. This is meant as a convenience and
        sanity check when you know the original file isn't supposed to exist.

        Note that 'x' is not suitable for preventing concurrent creation. This is
        due to the "eventually consistent" nature of s3, and also the CachedFile
        design. If two processes simultaneously write to the same file, there's a
        race condition. One or both will appear to succeed, and one will win out.
        Using 'x' doesn't -- can't -- prevent this.

        Raises s3.KeyNotFound when attempting to open a local or remote file that
        doesn't exist, s3.KeyExists when attempting exclusive create on an
        existing file or key, ValueError for an invalid key, and IOError for an
        underlying local file-system failure.

        '''
        return CachedFile(key, mode, connection=self)

    def list_buckets(self):
        '''
        List all the buckets availiable on AWS S3 for the credentialed account.
        '''
        return [x.name for x in self.conn.get_all_buckets()]

    def bucket_info(self, name):
        '''
        Get info about a bucket
        '''
        bucket = self._bucket(name)
        def safe_get(method):
            from boto import exception
            try:
                return getattr(bucket, method)()
            except exception.S3ResponseError:
                return None
        return {
            'name': name,
            'lifecycle': safe_get('get_lifecycle_config'),
            'cors': safe_get('get_cors'),
            'location': safe_get('get_location'),
            'logging_status': safe_get('get_logging_status'),
            'policy': safe_get('get_policy'),
            'versioning_status': safe_get('get_versioning_status'),
            'website_configuration': safe_get('get_website_configuration'),
            'website_endpoint': safe_get('get_website_endpoint'),
        }

    def create_bucket(self, name):
        return self.conn.create_bucket(name)


class FileMode(object):
    def __init__(self, mode, allowed_modes=None):
        self.mode = mode
        self._modes = set(mode)
        if self._modes - set('arwxb+tU') or len(self.mode) > len(self._modes):
            raise ValueError('Invalid mode')
        if allowed_modes:
            if self._modes - set(allowed_modes):
                raise NotImplementedError('CachedFile does not support mode %s' % "".join(self._modes - set(allowed_modes)))

        if 'U' in self._modes:
            raise NotImplementedError('s3.open does not support universal newline mode')
        self.reading = 'r' in self._modes
        self.writing = 'w' in self._modes
        self.creating_exclusively = 'x' in self._modes
        self.appending = 'a' in self._modes
        self.updating = '+' in self._modes
        self.text = 't' in self._modes
        self.binary = 'b' in self._modes
        if self.text and self.binary:
            raise ValueError("can't have text and binary mode at once")
        if self.reading + self.writing + self.appending + self.creating_exclusively > 1:
            raise ValueError("can't have read/write/append/exclusive create mode at once")
        if not (self.reading or self.writing or self.appending or self.creating_exclusively):
            raise ValueError("must have exactly one of read/write/append/exclusive create mode")

    @property
    def is_output(self):
        return self.writing or self.appending or self.creating_exclusively

    @property
    def flags(self):
        '''
        Adapted from http://hg.python.org/cpython/file/84cf25da86e8/Lib/_pyio.py#l154

        See also open(2) which explains the modes

        os.O_BINARY and os.O_TEXT are only available on Windows.
        '''
        return (
            ((self.reading and not self.updating) and os.O_RDONLY or 0) |
            ((self.writing and not self.updating) and os.O_WRONLY or 0) |
            ((self.creating_exclusively and not self.updating) and os.O_EXCL or 0) |
            (self.updating and os.O_RDWR or 0) |
            (self.appending and os.O_APPEND or 0) |
            ((self.writing or self.creating_exclusively) and os.O_CREAT or 0) |
            (self.writing and os.O_TRUNC or 0) |
            ((self.binary and hasattr(os, 'O_BINARY')) and os.O_BINARY or 0) |
            ((self.text and hasattr(os, 'O_TEXT')) and os.O_TEXT or 0)
        )

class CachedFile(object):
    '''
    CachedFile('s3://bucket/path/to/file.ext', 'r') downloads the file and opens it for reading
    CachedFile('s3://bucket/path/to/file.ext', 'w') opens a temp file for writing and uploads it on close
    CachedFile('s3://bucket/path/to/file.ext', 'x') verifies that the file doesn't exist on s3, then behaves like 'w'
    '''
    def __init__(self, key, mode='r', connection=None, encrypt=True):
        self.encrypt = encrypt
        self.key = key
        if path.islocal(key):
            self.should_upload_on_close = False
            self.mode = FileMode(mode, allowed_modes='arwxb+t')
            import __builtin__
            local_path = path.parse(key).path
            if self.mode.is_output and not os.path.exists(os.path.dirname(local_path)):
                from baiji.util.shutillib import mkdir_p
                mkdir_p(os.path.dirname(local_path))
            try:
                # Use os.open to catch exclusive access to the file, but use open to get a nice, useful file object
                self.fd = os.open(local_path, self.mode.flags)
                self.f = __builtin__.open(local_path, self.mode.mode.replace('x', 'w'))
                os.close(self.fd)
            except OSError as e:
                import errno
                if e.errno is errno.EEXIST:
                    raise KeyExists("Local file exists: %s" % local_path)
                elif e.errno is errno.ENOENT:
                    raise KeyNotFound("Local file does not exist: %s" % local_path)
                else:
                    raise IOError(e.errno, "%s: %s" % (e.strerror, e.filename))
        else:
            if connection is None:
                connection = S3Connection()
            self.connection = connection

            self.mode = FileMode(mode, allowed_modes='rwxbt')
            self.should_upload_on_close = self.mode.is_output
            if self.mode.creating_exclusively:
                if self.connection.exists(self.key):
                    raise KeyExists("Key exists in bucket: %s" % self.key)
                else:
                    self.connection.touch(self.key, encrypt=self.encrypt)
            # Use w+ so we can read back the contents in upload()
            new_mode = (
                'w+' +
                (self.mode.binary and 'b' or '') +
                (self.mode.text and 't' or '')
            )
            from baiji.util import tempfile
            self.f = tempfile.NamedTemporaryFile(
                mode=new_mode,
                suffix=os.path.splitext(path.parse(self.key).path)[1]
            )
            self.name = self.f.name
            self.remotename = key # Used by some serialization code to find files which sit along side the file in question, like textures which sit next to a mesh file
            if self.mode.reading:
                self.connection.cp(self.key, self.name, force=True)
    def upload(self):
        self.connection.cp(self.name, self.key, encrypt=self.encrypt, force=True)
    def __enter__(self):
        return self
    def __exit__(self, exc_type, exc_value, traceback):
        # If an unhandled exception is bubbling up, don't write to s3!
        if exc_value:
            self.should_upload_on_close = False
        self.close()
    def __getattr__(self, attr):
        return getattr(self.f, attr)
    def __iter__(self):
        return iter(self.f)
    def flush(self):
        self.f.flush()
        if self.should_upload_on_close:
            self.upload()
    def close(self):
        self.flush()
        self.f.close()


class path(object):
    def __init__(self):
        raise NotImplementedError("+++ Out of cheese error, redo from start +++")
    @classmethod
    def parse(cls, s):
        '''
        Parse a path given as a url. Accepts strings of the form:

           s3://bucket-name/path/to/key
           file:///path/to/file
           /absolution/path/to/file
           relative/path/to/file
           ~/path/from/home/dir/to/file

           To avoid surprises, s3:// and file:// URLs should not
           include ;, ? or #. You should URL-encode such paths.

        Return value is a ParseResult; one of the following:

           ('s3', bucketname, valid_s3_key, ...)
           ('file', '', absolute_path_for_current_filesystem, ...)

        '''
        from urlparse import urlparse, ParseResult

        if not isinstance(s, basestring):
            raise ValueError("An S3 path must be a string, got %s" % s.__class__.__name__)

        is_windows_path = (len(s) >= 2 and s[1] == ':')
        if is_windows_path:
            scheme, netloc, s3path = 'file', '', s
        else:
            scheme, netloc, s3path, params, query, fragment = urlparse(s)
            if any([params, query, fragment]):
                raise ValueError("Invalid URI: %s" % s)
            if any(char in ';?#' for char in s):
                raise ValueError("Invalid URI: %s" % s)
            try:
                s3path.encode('UTF-8')
            except (UnicodeDecodeError, UnicodeEncodeError):
                raise ValueError("Invalid URI (bad unicode): %s" % s)
                # If somehow something ever gets uploaded with binary in the
                # key, this seems to be the only way to fix it:
                # `s3cmd fixbucket s3://bodylabs-korper-assets`
        if re.match(r'/\w:', s3path): # urlparse, given file:///C:\foo parses us to /C:\foo, so on reconstruction (on windows) we get C:\C:\foo.
            s3path = s3path[1:]
            is_windows_path = True
        if scheme == '':
            scheme = 'file'
        if scheme == 'file' and not is_windows_path:
            if s3path.endswith(os.sep) or s3path.endswith('/'):
                # os.path.abspath strips the trailing '/' so we need to put it back
                s3path = os.path.join(os.path.abspath(os.path.expanduser(s3path)), '')
            else:
                s3path = os.path.abspath(os.path.expanduser(s3path))
        if scheme == 's3' and netloc == '':
            raise ValueError('s3 urls must specify the bucket')
        return ParseResult(scheme, netloc, s3path, params=None, query=None, fragment=None) # pylint: disable=too-many-function-args,unexpected-keyword-arg
    @classmethod
    def islocal(cls, s):
        '''
        Check if a path is local. Just parses and checks format, _does not_ check for existence of the file.
        '''
        return path.parse(s).scheme == 'file'
    @classmethod
    def isremote(cls, s):
        '''
        Check if a path is on S3. Just parses and checks format, _does not_ check for existence of the file.
        '''
        return path.parse(s).scheme == 's3'
    @classmethod
    def gettmpdir(cls, prefix='tmp/', suffix='', bucket='bodylabs-temp', uuid_generator=None):
        '''
        Make a directory on S3 with a known unique name. The prefix for the directory will be
        ``s3://<bucket>/<prefix><UUID><suffix>/``

        Note that there _is_ a race condition in this code; if two clients happen to be trying to create a tmpdir and
        somehow come up with the same uuid exactly simultaniously, they could both get the same dir. But in practice this
        is sufficiently one in a billion that we'll not worry about it for now.
        '''
        from boto.s3.key import Key
        if uuid_generator is None:
            from uuid import uuid4 as uuid_generator # pragma: no cover

        #s3.ls and s3.path.parse generated one with leading slash so remove it in case
        prefix = prefix.lstrip('/')
        b = S3Connection()._bucket(bucket) # pylint: disable=protected-access
        done = False
        while not done:
            tmppath = "%s%s%s" % (prefix, uuid_generator(), suffix)
            if len([x for x in b.list(prefix=tmppath)]) == 0:
                k = Key(b)
                k.key = "%s/.tempdir" % (tmppath)
                k.set_contents_from_string('')
                done = True
        return "s3://%s/%s/" % (bucket, tmppath)

    @classmethod
    def join(cls, base, *additions):
        '''
        Extends os.path.join so work with s3:// and file:// urls

        This inherits a quirk of os.path.join: if 'addition' is
        an absolute path, path components of base are thrown away.

        'addition' must be an absolute or relative path, not
        a URL.

        `base` and `addition` can use any path separator, but the
        result will always be normalized to os.sep.

        '''
        from urlparse import urlparse, urljoin, ParseResult

        addition = sep.join(additions)

        (scheme, netloc, _, params, query, fragment) = urlparse(addition)
        if any([scheme, netloc, params, query, fragment]):
            raise ValueError('Addition must be an absolute or relative path, not a URL')

        if cls.islocal(base):
            return os.path.join(cls.parse(base).path, addition.replace(sep, os.sep))
        k = cls.parse(base)

        # Call urljoin instead of os.path.join, since it uses '/' instead of
        # os.sep, which is '\' on Windows.
        #
        # Given disparity between os.path.join and urljoin, we prefer the
        # behavior of os.path.join:
        #
        #   >>> os.path.join('foo/bar', 'baz')
        #   'foo/bar/baz'
        #   >>> urlparse.urljoin('foo/bar', 'baz')
        #   'foo/baz'
        #
        # So we add a trailing slash if there is none
        if k.path.endswith(sep):
            s3path = urljoin(k.path, addition)
        else:
            s3path = urljoin(k.path + sep, addition)

        return ParseResult(k.scheme, k.netloc, s3path, k.params, k.query, k.fragment).geturl() # pylint: disable=too-many-function-args,unexpected-keyword-arg

    @classmethod
    def basename(cls, key):
        '''
        Finds the basename of a file on s3 or local. For local, it's equivalent to os.path.basename
        '''
        k = cls.parse(key)
        return os.path.basename(k.path)

    @classmethod
    def dirname(cls, key):
        '''
        Finds the dirname of a file on s3 or local. For local, it's equivalent to os.path.dirname
        '''
        # Oddly enough, os.path.dirname works correctly on URIs...
        return os.path.dirname(key)

    @classmethod
    def isdirlike(cls, key):
        """
        Returns True for any key that is either a local, existing directory or
        ends with `sep` or `os.sep`. Otherwise returns False.
        This preserves the old `isdir` behavior.
        """
        k = cls.parse(key)
        if cls.islocal(key) and os.path.isdir(k.path):
            return True
        else:
            return k.path.endswith(sep) or k.path.endswith(os.sep)

    @classmethod
    def isdir(cls, key):
        '''
        Return true if key is directory-ish. That is, it ends with a path
        separator, or is a local directory that actually exists.
        On S3 a "directory" is considered to exist if one or more files exist
        that have the "directory" (ending with sep) as a prefix.
        '''
        k = cls.parse(key)
        if cls.islocal(key): #This really only ensures that scheme == 'file'
            return os.path.isdir(k.path)
        if cls.isremote(key): # scheme == 'S3'
            if not k.path.endswith(sep):
                k = cls.parse(key + sep)
            try:
                next(ls(k.geturl()))
                return True
            except StopIteration:
                return False
        else:
            raise InvalidSchemeException("URI Scheme {} is not implemented".format(k.scheme))

    @classmethod
    def isfile(cls, key):
        '''
        Return true if key is file; local or s3.
        '''
        k = cls.parse(key)
        if cls.islocal(key): #This really only ensures that scheme == 'file'
            return os.path.isfile(k.path)
        if cls.isremote(key): # scheme == 'S3'
            # exists currently only works for files on s3 because
            # directories don't exist on s3, only files.
            return exists(key)
        else:
            raise InvalidSchemeException("URI Scheme {} is not implemented".format(k.scheme))

    @classmethod
    def bucket(cls, key):
        '''
        Extracts the bucket from a key.
        '''
        k = cls.parse(key)
        return k.netloc



import unittest
import os

class TestAWS(unittest.TestCase):

    def test_credentials(self):
        from baiji.config import credentials
        from baiji.util import yaml
        bodylabs_file_path = os.getenv('BODYLABS_CREDENTIAL_FILE', os.path.expanduser('~/.bodylabs'))
        if not os.path.exists(bodylabs_file_path):
            raise unittest.SkipTest("Skipping test_credentials because ~/.bodylabs doesn't exist.")
        truth = yaml.load(bodylabs_file_path)
        self.assertEqual(credentials.key, truth['AWS_ACCESS_KEY'])
        self.assertEqual(credentials.secret, truth['AWS_SECRET'])



'''
tempfile does not seem to support `tempdir` correctly on all platforms,
so we add an override variable that we always support.

If the environment variable BAIJI_TMP is set, baiji will always write
its temporary download files there.
'''
from __future__ import absolute_import
import os
from baiji.util.shutillib import mkdir_p

def _override_tmp(fn, args, kwargs):
    if 'dir' not in kwargs:
        kwargs['dir'] = os.environ.get('BAIJI_TMP', None)
        if kwargs['dir'] is not None:
            mkdir_p(kwargs['dir'])
    return fn(*args, **kwargs)


def TemporaryFile(*args, **kwargs):
    from tempfile import TemporaryFile as _TemporaryFile
    return _override_tmp(_TemporaryFile, args, kwargs)

def NamedTemporaryFile(*args, **kwargs):
    from tempfile import NamedTemporaryFile as _NamedTemporaryFile
    return _override_tmp(_NamedTemporaryFile, args, kwargs)

def mkstemp(*args, **kwargs):
    from tempfile import mkstemp as _mkstemp
    return _override_tmp(_mkstemp, args, kwargs)

def mkdtemp(*args, **kwargs):
    from tempfile import mkdtemp as _mkdtemp
    return _override_tmp(_mkdtemp, args, kwargs)



import unittest

class TestMD5(unittest.TestCase):
    def test_md5(self):
        import os
        import struct
        from baiji.util import tempfile
        from baiji.util.md5 import md5_for_file
        f = tempfile.NamedTemporaryFile('wb', delete=False)
        for _ in range(2**16):
            f.write(struct.pack("<I", 42))
        f.close()
        self.assertEqual(md5_for_file(f.name), "9b0cbf6ba5c0ad5606df9b9630a44db6")
        os.remove(f.name)



import os
import shutil
import unittest
import mock
from baiji import s3
from baiji.util.testing import create_random_temporary_file
from baiji.util import tempfile

TEST_BUCKET = 'baiji-test'

class TestAWSBase(unittest.TestCase):
    """
    Provides framework for testing.
    Among others:
    self.local_file                         # A local file that exists
    self.tmp_dir                            # A local temporary directory (that exists)
    self.existing_remote_file               # A remote file that exists
    self.s3_path                            # A remote temporary "directory" (N.B. it does not "exist" a priori)
                                            # Note: this should be accessed with self.remote_file(filename) for full s3 url
    self.retriable_s3_call(call, retries=3) # For testing methods that may fail due to bad connections
    """
    def setUp(self):
        self.bucket = TEST_BUCKET
        self.s3_test_location = s3.path.gettmpdir(bucket=self.bucket)
        loc = s3.path.parse(self.s3_test_location)
        self.s3_path = loc.path

        b = s3.S3Connection()._bucket(self.bucket) # pylint: disable=protected-access
        for key in b.list(self.s3_path[1:]):
            b.delete_key(key)
        self.assertEqual(len([x for x in b.list(self.s3_path[1:])]), 0) # just make sure we're starting with a clean slate

        self.tmp_dir = tempfile.mkdtemp('bodylabs-test')
        self.local_file = create_random_temporary_file()


    def tearDown(self):
        shutil.rmtree(self.tmp_dir, ignore_errors=True)
        os.remove(self.local_file)
        b = s3.S3Connection()._bucket(self.bucket) # pylint: disable=protected-access
        for key in b.list(self.s3_path[1:]):
            b.delete_key(key)
        self.assertEqual(len([x for x in b.list(self.s3_path[1:])]), 0)

    def retriable_s3_call(self, call, retries=3):
        from boto.exception import S3ResponseError
        import time
        try:
            return call()
        except S3ResponseError:
            if retries > 0:
                time.sleep(0.1)
                return self.retriable_s3_call(call, retries=retries-1)
            else:
                raise

    def assert_s3_exists(self, path):
        self.assertTrue(self.retriable_s3_call(lambda: s3.exists(path)))

    def assert_s3_does_not_exist(self, path):
        self.assertFalse(self.retriable_s3_call(lambda: s3.exists(path)))

    def assert_is_public(self, s3_url, is_public):
        from urlparse import urlparse
        url = urlparse(s3_url)
        acl = self.retriable_s3_call(lambda: s3.S3Connection()._bucket(url.netloc).get_acl(url.path[1:])) # pylint: disable=protected-access
        actual_is_public = False
        for grant in acl.acl.grants:
            if grant.permission == 'READ':
                if grant.uri == 'http://acs.amazonaws.com/groups/global/AllUsers':
                    actual_is_public = True
        self.assertEquals(actual_is_public, is_public)

    def remote_file(self, path):
        return s3.path.join(self.s3_test_location, path)

    @property
    def existing_remote_file(self):
        '''
        In some tests it is convenient to have a file already on s3;
        in some others we need it not to be there (e.g. for clarity in the s3.ls test)
        '''
        uri = self.remote_file("FOO/A_preexisting_file.md")
        if not s3.exists(uri):
            s3.cp(self.local_file, uri)
        return uri

class TestS3Exists(TestAWSBase):

    @mock.patch('baiji.s3.S3Connection._lookup')
    def test_s3_exists_retries_if_not_found_at_first(self, mock_lookup):
        import warnings
        from baiji.exceptions import EventualConsistencyWarning
        mock_key = "all_we_care_is_that_it's not None"
        mock_lookup.side_effect = [None, None, mock_key]
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            self.assertTrue(s3.exists('s3://foo'))
            # Verify the warning was triggered
            self.assertEqual(len(w), 1)
            self.assertIs(w[-1].category, EventualConsistencyWarning)
            self.assertEqual(str(w[-1].message), "S3 is behaving in an eventually consistent way in s3.exists(s3://foo) -- it took 3 attempts to locate the key")
        self.assertEqual(mock_lookup.call_count, 3)

    @mock.patch('baiji.s3.S3Connection._lookup')
    def test_s3_exists_does_not_retry_if_found_immidiately(self, mock_lookup):
        mock_key = "all_we_care_is_that_it's not None"
        mock_lookup.return_value = mock_key
        self.assertTrue(s3.exists('s3://foo'))
        self.assertEqual(mock_lookup.call_count, 1)

    @mock.patch('baiji.s3.S3Connection._lookup')
    def test_s3_exists_return_false_if_the_file_never_shows_up(self, mock_lookup):
        mock_lookup.return_value = None
        self.assertFalse(s3.exists('s3://foo'))
        self.assertEqual(mock_lookup.call_count, 3)


class TestS3(TestAWSBase):

    def test_s3_cp_local_to_local(self):
        s3.cp(self.local_file, os.path.join(self.tmp_dir, 'TEST.foo'))
        self.assertTrue(os.path.exists(os.path.join(self.tmp_dir, 'TEST.foo')))

    def test_s3_cp_local_to_local_file_in_dir_that_needs_making(self):
        s3.cp(self.local_file, os.path.join(self.tmp_dir, 'FOO', 'TEST.foo'))
        self.assertTrue(os.path.exists(os.path.join(self.tmp_dir, 'FOO', 'TEST.foo')))

    def test_s3_cp_local_to_local_dir(self):
        s3.cp(self.local_file, self.tmp_dir)
        self.assertTrue(os.path.exists(os.path.join(self.tmp_dir, os.path.basename(self.local_file))))

    def test_s3_cp_upload(self):
        s3.cp(self.local_file, self.remote_file("FOO/NAMED.md"))
        self.assert_s3_exists(self.remote_file("FOO/NAMED.md"))

    def test_s3_cp_download(self):
        s3.cp(self.existing_remote_file, os.path.join(self.tmp_dir, 'DL', 'TEST.foo'))
        self.assertTrue(os.path.exists(os.path.join(self.tmp_dir, 'DL', 'TEST.foo')))
        s3.cp(self.existing_remote_file, os.path.join(self.tmp_dir, 'DL'))
        self.assertTrue(os.path.exists(os.path.join(self.tmp_dir, 'DL', s3.path.basename(self.existing_remote_file))))

    @mock.patch('baiji.s3.S3CopyOperation.ensure_integrity')
    def test_s3_cp_download_corrupted_recover_in_one_retry(self, ensure_integrity_mock):
        from baiji.s3 import get_transient_error_class
        ensure_integrity_mock.side_effect = [get_transient_error_class()('etag does not match'), None]
        s3.cp(self.existing_remote_file, self.tmp_dir, force=True)

    @mock.patch('boto.s3.key.Key.get_contents_to_file')
    def test_downloads_from_s3_are_atomic_under_truncation(self, download_mock):
        from baiji.s3 import get_transient_error_class
        def write_fake_truncated_file(fp, **kwargs): # just capturing whatever is thrown at us: pylint: disable=unused-argument
            fp.write("12345")
        download_mock.side_effect = write_fake_truncated_file
        # Now when the call to download the file is made, the etags won't match
        with self.assertRaises(get_transient_error_class()):
            s3.cp(self.existing_remote_file, os.path.join(self.tmp_dir, 'truncated.foo'), validate=True)
        self.assertFalse(os.path.exists(os.path.join(self.tmp_dir, 'truncated.foo')))

    @mock.patch('baiji.s3.S3CopyOperation.ensure_integrity')
    def test_downloads_from_s3_are_atomic_under_exceptions(self, download_mock):
        download_mock.side_effect = ValueError()
        # Now when the call to download the file is made, an exception will be thrown.
        # ideally, we'd throw it "in" boto via a mock, but we really want to test that
        # the file doesn't get written, so let's go ahead and let boto do the download
        # and then throw the exception in the validation
        with self.assertRaises(ValueError):
            s3.cp(self.existing_remote_file, os.path.join(self.tmp_dir, 'erroneous.foo'), validate=True)
        self.assertFalse(os.path.exists(os.path.join(self.tmp_dir, 'erroneous.foo')))

    @mock.patch('baiji.s3.S3CopyOperation.ensure_integrity')
    def test_s3_cp_download_corrupted_raise_transient_error_after_retried_once(self, ensure_integrity_mock):

        from baiji.s3 import get_transient_error_class

        ensure_integrity_mock.side_effect = get_transient_error_class()('etag does not match')

        with self.assertRaises(get_transient_error_class()):
            s3.cp(self.existing_remote_file, self.tmp_dir, force=True)

    def test_s3_cp_in_s3(self):
        s3.cp(self.existing_remote_file, self.remote_file("COPY/NAMED.md"))
        self.assert_s3_exists(self.remote_file("COPY/NAMED.md"))

    def test_s3_cp_overwrite_errors_unless_force(self):
        s3.cp(self.local_file, os.path.join(self.tmp_dir, os.path.basename(self.local_file)))
        # local to local
        with self.assertRaises(s3.KeyExists):
            s3.cp(self.local_file, os.path.join(self.tmp_dir, os.path.basename(self.local_file)))
        with self.assertRaises(s3.KeyExists):
            s3.cp(self.local_file, self.tmp_dir)
        s3.cp(self.local_file, os.path.join(self.tmp_dir, os.path.basename(self.local_file)), force=True)
        s3.cp(self.local_file, self.tmp_dir, force=True)
        # local to remote
        with self.assertRaises(s3.KeyExists):
            s3.cp(self.local_file, self.existing_remote_file)
        # remote to local
        with self.assertRaises(s3.KeyExists):
            s3.cp(self.existing_remote_file, self.local_file)
        # remote to remote
        with self.assertRaises(s3.KeyExists):
            s3.cp(self.existing_remote_file, self.existing_remote_file)

    def test_s3_cp_errors_if_source_is_missing(self):
        # remote
        with self.assertRaisesRegexp(s3.KeyNotFound, "Error copying"):
            s3.cp(self.remote_file("definitely_not_there.foo"), self.tmp_dir)
        # local
        with self.assertRaisesRegexp(s3.KeyNotFound, "Error copying"):
            s3.cp(os.path.join(self.tmp_dir, "definitely_not_there.foo"), self.tmp_dir)


    def test_s3_cp_skip_existing_files_without_raise_exceptions(self):
        import warnings
        s3.cp(self.local_file, os.path.join(self.tmp_dir, os.path.basename(self.local_file)))
        with self.assertRaises(s3.KeyExists):
            s3.cp(self.local_file, os.path.join(self.tmp_dir, os.path.basename(self.local_file)))
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            s3.cp(self.local_file, os.path.join(self.tmp_dir, os.path.basename(self.local_file)), skip=True)
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)
            assert "Skipping" in str(w[-1].message)

    def test_s3_cp_errors_raised_by_missing_source_file(self):
        self.assertRaises(s3.KeyNotFound, s3.cp, os.path.join(self.tmp_dir, 'MISSING.file'), self.tmp_dir)
        self.assertRaises(s3.KeyNotFound, s3.cp, self.remote_file("MISSING.file"), self.tmp_dir)

    def test_s3_cp_errors_raised_by_unimplemented_source_scheme(self):
        self.assertRaises(s3.InvalidSchemeException, s3.cp, 'http://www.google.com/', self.tmp_dir)
        self.assertRaises(s3.InvalidSchemeException, s3.cp, self.local_file, 'http://www.google.com/')

    def test_s3_cp_relative_paths(self):
        s3.cp(self.local_file, 'test-DELETEME', force=True)
        self.assertTrue(os.path.exists(os.path.join(os.getcwd(), 'test-DELETEME')))
        s3.cp('test-DELETEME', self.remote_file("local"), force=True)
        self.assert_s3_exists(self.remote_file("local"))
        s3.rm('test-DELETEME')
        self.assertFalse(os.path.exists(os.path.join(os.getcwd(), 'test-DELETEME')))

    def test_s3_cp_HOME_paths(self):
        s3.cp(self.local_file, '~/test-DELETEME', force=True)
        self.assertTrue(os.path.exists(os.path.expanduser('~/test-DELETEME')))
        s3.cp('~/test-DELETEME', self.remote_file("local"), force=True)
        self.assert_s3_exists(self.remote_file("local"))
        s3.rm('~/test-DELETEME')
        self.assertFalse(os.path.exists(os.path.expanduser('~/test-DELETEME')))

    def test_s3_cp_policy(self):
        # test policy for file -> s3
        s3.cp(self.local_file, self.remote_file("public.md"), policy='public-read')
        self.assert_is_public(self.remote_file("public.md"), is_public=True)
        s3.cp(self.local_file, self.remote_file("private.md"), policy='bucket-owner-read')
        self.assert_is_public(self.remote_file("private.md"), is_public=False)

        # test policy for s3 -> s3
        s3.cp(self.remote_file("private.md"), self.remote_file("made_public_on_copy.md"), policy='public-read')
        self.assert_is_public(self.remote_file("made_public_on_copy.md"), is_public=True)

        s3.cp(self.remote_file("private.md"), self.remote_file("left_private_on_copy.md"))
        self.assert_is_public(self.remote_file("left_private_on_copy.md"), is_public=False)

        with self.assertRaises(ValueError):
            s3.cp(self.remote_file("private.md"), os.path.join(self.tmp_dir, 'NoCanDo.txt'), policy='public-read')

    def test_s3_cp_preserve_acl(self):
        s3.cp(self.local_file, self.remote_file("also_public.md"), policy='public-read')
        s3.cp(self.remote_file("also_public.md"), self.remote_file("still_public.md"), preserve_acl=True)
        self.assert_is_public(self.remote_file("also_public.md"), is_public=True)
        s3.cp(self.remote_file("also_public.md"), self.remote_file("no_longer_public.md"))
        self.assert_is_public(self.remote_file("no_longer_public.md"), is_public=False)

        with self.assertRaises(ValueError):
            s3.cp(self.remote_file("also_public.md"), os.path.join(self.tmp_dir, 'NoCanDo.txt'), preserve_acl=True)

    def test_s3_cp_content_encoding(self):
        s3.cp(self.local_file, self.remote_file("encoded.md"), encoding='gzip')
        self.assertEqual(s3.info(self.remote_file("encoded.md"))['content_encoding'], 'gzip')
        s3.cp(self.local_file, self.remote_file("notencoded.md")) # just make sure gzip isn't the default ;)
        self.assertNotEqual(s3.info(self.remote_file("notencoded.md"))['content_encoding'], 'gzip')
        s3.cp(self.remote_file("encoded.md"), self.remote_file("still_encoded.md"))
        self.assertEqual(s3.info(self.remote_file("still_encoded.md"))['content_encoding'], 'gzip')
        s3.cp(self.remote_file("notencoded.md"), self.remote_file("now_encoded.md"), encoding='gzip')
        self.assertEqual(s3.info(self.remote_file("now_encoded.md"))['content_encoding'], 'gzip')

    def test_s3_cp_content_type(self):
        s3.cp(self.local_file, self.remote_file("typed.md"), content_type='text/html')
        self.assertEqual(s3.info(self.remote_file("typed.md"))['content_type'], 'text/html')

        s3.cp(self.local_file, self.remote_file("nottyped.md")) # default is
        self.assertEqual(s3.info(self.remote_file("nottyped.md"))['content_type'], 'application/octet-stream')

        s3.cp(self.remote_file("typed.md"), self.remote_file("still_typed.md"))
        self.assertEqual(s3.info(self.remote_file("still_typed.md"))['content_type'], 'text/html')
        s3.cp(self.remote_file("nottyped.md"), self.remote_file("now_typed.md"), content_type='text/html')
        self.assertEqual(s3.info(self.remote_file("now_typed.md"))['content_type'], 'text/html')

    def test_s3_cp_gzip(self):
        s3.cp(self.local_file, self.remote_file("big.md"))
        s3.cp(self.local_file, self.remote_file("small.md"), gzip=True)
        self.assertNotEqual(s3.info(self.remote_file("big.md"))['content_encoding'], 'gzip')
        self.assertEqual(s3.info(self.remote_file("small.md"))['content_encoding'], 'gzip')
        self.assertLess(s3.info(self.remote_file("small.md"))['size'], s3.info(self.remote_file("big.md"))['size'])

    def test_s3_cp_trailing_slashes_in_dst(self):
        # https://bitbucket.org/bodylabs/core/issue/25
        s3.cp(self.local_file, os.path.join(self.tmp_dir, 'FOO2', ''))
        self.assertTrue(os.path.exists(os.path.join(self.tmp_dir, 'FOO2', os.path.basename(self.local_file))))
        s3.cp(self.local_file, self.remote_file("FOO2/"))
        self.assert_s3_exists(self.remote_file("FOO2/%s" % os.path.basename(self.local_file)))
        s3.cp(self.existing_remote_file, os.path.join(self.tmp_dir, 'DL2', ''))
        self.assertTrue(os.path.exists(os.path.join(self.tmp_dir, 'DL2', s3.path.basename(self.existing_remote_file))))

    def test_s3_rm(self):
        for path in [os.path.join(self.tmp_dir, 'foo'), self.remote_file("foo")]:
            s3.cp(self.local_file, path)
            self.assert_s3_exists(path)
            s3.rm(path)
            self.assert_s3_does_not_exist(path)
            with self.assertRaises(s3.KeyNotFound):
                s3.rm(path)
        self.assertRaises(s3.InvalidSchemeException, s3.rm, ("http://example.com/foo"))

    def test_s3_ls(self):
        files = ["foo", "bar.baz", "quack/foo.foo"]
        for f in files:
            s3.cp(self.local_file, self.remote_file(f))
        self.assertEqual(set(s3.ls(self.s3_test_location)), set(map(lambda x: self.s3_path+x, files)))
        self.assertEqual(set(s3.ls(self.s3_test_location, return_full_urls=True)), set([self.remote_file(x) for x in files]))
        self.assertEqual(set(s3.ls(self.s3_test_location, shallow=True)), set([self.s3_path+x for x in ['foo', 'bar.baz', 'quack/']]))

    def test_s3_md5(self):
        s3.cp(self.local_file, self.remote_file("file_to_md5"))
        from baiji.util.md5 import md5_for_file
        self.assertEqual(s3.md5(self.remote_file("file_to_md5")), md5_for_file(self.local_file))

    def test_strings(self):
        s = "TEST STRING"
        s3.put_string(self.remote_file("string"), s)
        self.assertEqual(s3.get_string(self.remote_file("string")), s)

    # TODO: Test with local paths. Instead of having a single local file
    # we can have a local subdirectory that is cleaned up on teardown.
    def test_s3_glob_match_single_wildcard(self):
        files = ['a.obj', 'b.obj', 'a.ply', 'a.object']
        for f in files:
            s3.cp(self.local_file, self.remote_file(f))
        glob_results = list(s3.glob(self.s3_test_location, '*.obj'))
        self.assertEqual(2, len(glob_results))
        self.assertEqual(set(glob_results), set([self.remote_file(f) for f in ['a.obj', 'b.obj']]))

    def test_s3_glob_match_multiple_wildcards(self):
        files = ['body_1_pose_T.obj', 'body_1_pose_Fun.obj', 'body_2_pose_T.obj', 'body_02_pose_T.obj']
        for f in files:
            s3.cp(self.local_file, self.remote_file(f))
        glob_results = list(s3.glob(self.s3_test_location, 'body_?_pose_*.obj'))
        self.assertEqual(3, len(glob_results))
        self.assertEqual(
            set(glob_results),
            set([self.remote_file(f) for f in ['body_1_pose_T.obj', 'body_1_pose_Fun.obj', 'body_2_pose_T.obj']])
        )

    def test_s3_glob_exclude(self):
        files = ['pose_T.obj', 'pose_A.obj', 'pose_Scan.obj']
        for f in files:
            s3.cp(self.local_file, self.remote_file(f))
        glob_results = list(s3.glob(self.s3_test_location, 'pose_[!T].obj'))
        self.assertEqual(1, len(glob_results))
        self.assertEqual(set(glob_results), set([self.remote_file('pose_A.obj')]))

    def test_s3_with_double_slashes_in_key(self):
        '''
        boto has a nasty behavior by default where it collapses `//` to `/` in keys
        '''
        s3.cp(self.local_file, self.remote_file('double//slashes//bork//boto.foo'))
        self.assertEqual([self.remote_file('double//slashes//bork//boto.foo')], list(s3.ls(self.remote_file(''), return_full_urls=True)))

    def test_s3_path_isdir(self):
        existing_remote_dir = os.path.dirname(self.existing_remote_file)
        #case: local dir, exists
        self.assertTrue(s3.path.isdir(self.tmp_dir))
        #case: local dir, does not exist
        self.assertFalse(s3.path.isdir(os.path.join(self.tmp_dir, 'does_not_exist')))
        #case: local dir, exists but is file
        self.assertFalse(s3.path.isdir(self.local_file))
        #case: remote dir, exists
        self.assertTrue(s3.path.isdir(existing_remote_dir))
        #case: remote dir, does not exist
        self.assertFalse(s3.path.isdir(self.remote_file('does_not_exist')))
        #case: remote dir, exists but is file
        self.assertFalse(s3.path.isdir(self.existing_remote_file))

    def test_s3_path_isfile(self):
        existing_remote_dir = os.path.dirname(self.existing_remote_file)
        #case: local file, exists
        self.assertTrue(s3.path.isfile(self.local_file))
        #case: local file, does not exist
        self.assertFalse(s3.path.isfile(os.path.join(self.tmp_dir, 'does_not_exist')))
        #case: local file, exists but is directory
        self.assertFalse(s3.path.isfile(self.tmp_dir))
        #case: remote file, exists
        self.assertTrue(s3.path.isfile(self.existing_remote_file))
        #case: remote file, does not exist
        self.assertFalse(s3.path.isfile(self.remote_file('does_not_exist')))
        #case: remote file, "exists" but is directory (so it actually doesn't exist but could return false positive)
        self.assertFalse(s3.path.isfile(existing_remote_dir))


class TestEncryption(TestAWSBase):
    def test_encrypt_in_place(self):
        s3.cp(self.local_file, self.remote_file("to_encrypt.txt"), encrypt=False) # just make sure there's something to copy
        self.assertFalse(s3.info(self.remote_file("to_encrypt.txt"))['encrypted'])
        s3.encrypt_at_rest(self.remote_file("to_encrypt.txt"))
        self.assertTrue(s3.info(self.remote_file("to_encrypt.txt"))['encrypted'])

    def test_upload(self):
        s3.cp(self.local_file, self.remote_file("unencrypted.txt"), encrypt=False)
        self.assertFalse(s3.info(self.remote_file("unencrypted.txt"))['encrypted'])
        s3.cp(self.local_file, self.remote_file("encrypted.txt")) # default now to encrypt
        self.assertTrue(s3.info(self.remote_file("encrypted.txt"))['encrypted'])

    def test_copy(self):
        s3.cp(self.local_file, self.remote_file("unencrypted.txt"), encrypt=False) # just make sure there's something to copy
        self.assertFalse(s3.info(self.remote_file("unencrypted.txt"))['encrypted'])
        s3.cp(self.remote_file("unencrypted.txt"), self.remote_file("encrypted.txt"))
        self.assertTrue(s3.info(self.remote_file("encrypted.txt"))['encrypted'])


class TestS3ConnectionPersistence(unittest.TestCase):
    def test_connection_is_cached(self):
        # pylint: disable=protected-access
        # Using a cached connection is 2 orders of magnitude faster than
        # creating a new connection instance.

        connection = s3.S3Connection()

        # Creating the object does not a connection make
        self.assertFalse(connection._connected)

        # connection.conn is created on first use and contains the connection object
        # Whenever s3 operations are performed, a call is made to this object.
        # In order to eliminate the middleman, we make the call directly in this test.
        established_connection_id = id(connection.conn)

        # Now it should be registered that the connection is made and cached.
        self.assertTrue(connection._connected)

        # And connection.conn should not change when accessed again.
        connection_id_on_second_call = id(connection.conn)
        self.assertEquals(established_connection_id, connection_id_on_second_call)
